# calibrate nonlin

import io
import os
import warnings
import numpy as np
import pandas as pd
from pathlib import Path
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
from statsmodels.nonparametric.smoothers_lowess import lowess

from corgidrp import check
import corgidrp.data as data
from corgidrp.calibrate_kgain import CalKgainException

# Dictionary with constant non-linearity calibration parameters
nonlin_params_default = {
    # ROI constants
    'rowroi1': 305,
    'rowroi2': 736,
    'colroi1': 1385,
    'colroi2': 1846,
     
    # background ROIs
    'rowback11': 20,
    'rowback12': 301,
    'rowback21': 740,
    'rowback22': 1001,
    'colback11': 1200,
    'colback12': 2001,
    'colback21': 1200,
    'colback22': 2001,
     
    # minimum exposure time, s
    'min_exp_time': 0,

    # histogram bin parameters; min_bin is in DN
    'num_bins': 50,
    'min_bin': 200,
     
    # factor to mutiply bin_edge values when making mask
    'min_mask_factor': 1.1,
    }
 
def check_nonlin_params(nonlin_params):
    """ Checks integrity of kgain parameters in the dictionary nonlin_params. 

    Args:
        nonlin_params (dict):  Dictionary of parameters used for calibrating nonlinearity.
    """
    if 'rowroi1' not in nonlin_params:
        raise ValueError('Missing parameter:  rowroi1.')
    if 'rowroi2' not in nonlin_params:
        raise ValueError('Missing parameter:  rowroi2.')
    if 'colroi1' not in nonlin_params:
        raise ValueError('Missing parameter:  colroi1.')
    if 'colroi2' not in nonlin_params:
        raise ValueError('Missing parameter:  colroi2.')
    if 'rowback11' not in nonlin_params:
        raise ValueError('Missing parameter:  rowback11.')
    if 'rowback12' not in nonlin_params:
        raise ValueError('Missing parameter:  rowback12.')
    if 'rowback21' not in nonlin_params:
        raise ValueError('Missing parameter:  rowback21.')
    if 'rowback22' not in nonlin_params:
        raise ValueError('Missing parameter:  rowback22.')
    if 'colback11' not in nonlin_params:
        raise ValueError('Missing parameter:  colback11.')
    if 'colback12' not in nonlin_params:
        raise ValueError('Missing parameter:  colback12.')
    if 'colback21' not in nonlin_params:
        raise ValueError('Missing parameter:  colback21.')
    if 'colback22' not in nonlin_params:
        raise ValueError('Missing parameter:  colback22.')
    if 'min_exp_time' not in nonlin_params:
        raise ValueError('Missing parameter:  min_exp_time.')
    if 'num_bins' not in nonlin_params:
        raise ValueError('Missing parameter:  num_bins.')
    if 'min_bin' not in nonlin_params:
        raise ValueError('Missing parameter:  min_bin.')
    if 'min_mask_factor' not in nonlin_params:
        raise ValueError('Missing parameter:  min_mask_factor.')
    
    if not isinstance(nonlin_params['rowroi1'], (float, int)):
        raise TypeError('rowroi1 is not a number')
    if not isinstance(nonlin_params['rowroi2'], (float, int)):
        raise TypeError('rowroi2 is not a number')
    if not isinstance(nonlin_params['colroi1'], (float, int)):
        raise TypeError('colroi1 is not a number')
    if not isinstance(nonlin_params['colroi2'], (float, int)):
        raise TypeError('colroi2 is not a number')
    if not isinstance(nonlin_params['rowback11'], (float, int)):
        raise TypeError('rowback11 is not a number')
    if not isinstance(nonlin_params['rowback12'], (float, int)):
        raise TypeError('rowback12 is not a number')
    if not isinstance(nonlin_params['rowback21'], (float, int)):
        raise TypeError('rowback21 is not a number')
    if not isinstance(nonlin_params['rowback22'], (float, int)):
        raise TypeError('rowback22 is not a number')
    if not isinstance(nonlin_params['colback11'], (float, int)):
        raise TypeError('colback11 is not a number')
    if not isinstance(nonlin_params['colback12'], (float, int)):
        raise TypeError('colback12 is not a number')
    if not isinstance(nonlin_params['colback21'], (float, int)):
        raise TypeError('colback21 is not a number')
    if not isinstance(nonlin_params['colback22'], (float, int)):
        raise TypeError('colback22 is not a number')
    if not isinstance(nonlin_params['min_exp_time'], (float, int)):
        raise TypeError('min_exp_time is not a number')
    if not isinstance(nonlin_params['num_bins'], (float, int)):
        raise TypeError('num_bins is not a number')
    if not isinstance(nonlin_params['min_bin'], (float, int)):
        raise TypeError('min_bin is not a number')
    if not isinstance(nonlin_params['min_mask_factor'], (float, int)):
        raise TypeError('min_mask_factor is not a number')
    

class CalNonlinException(Exception):
    """Exception class for calibrate_nonlin."""

def calibrate_nonlin(dataset_nl,
                     n_cal=20, n_mean=30, norm_val = 2500, min_write = 800.0,
                     max_write = 10000.0,
                     lowess_frac = 0.1, rms_low_limit = 0.004, rms_upp_limit = 0.2,
                     pfit_upp_cutoff1 = -2, pfit_upp_cutoff2 = -3,
                     pfit_low_cutoff1 = 2, pfit_low_cutoff2 = 1,
                     make_plot=True, plot_outdir='figures', show_plot=False,
                     verbose=False, nonlin_params=None):
    """
    Function that derives the non-linearity calibration table for a set of DN
    and EM values.

    Args:
      dataset_nl (corgidrp.Dataset): The frames in the dataset are
        bias-subtracted. The dataset contains frames belonging to two different
        sets -- Mean frame, a large array of unity gain frames, and set with
        non-unity gain frames.
        Mean frame -- Unity gain frames with constant exposure time. These frames
        are used to create a mean pupil image. The mean frame is used to select
        pixels in each frame of the large array of unity gain frames (see next)
        to calculate its mean signal. In general, it is expected that at least
        30 frames or more will be taken for this set. In TVAC, 30 frames, each
        with an exposure time of 5.0 sec were taken.
        Large array of unity gain frames: Set of unity gain frames with subsets
        of equal exposure times. Data for each subset should be taken sequentially:
        Each subset must have at least 5 frames. All frames for a subset are taken
        before moving to the next subset. Two of the subsets have the same (repeated)
        exposure time. These two subsets are not contiguous: The first subset is
        taken near the start of the data collection and the second one is taken
        at the end of the data collection (see TVAC example below). The mean
        signal of these two subsets is used to correct for illumination
        brightness/sensor sensitivity drifts for all the frames in the whole set,
        depending on when the frames were taken. There should be no other repeated
        exposure time among the subsets. In TVAC, a total of 110 frames were taken
        within this category. The 110 frames consisted of 22 subsets, each with
        5 frames. All 5 frames had the same exposure time. The exposure times in
        TVAC in seconds were, each repeated 5 times to collect 5 frames in each
        subset -- 0.077, 0.770, 1.538, 2.308, 3.077, 3.846, 4.615, 5.385, 6.154,
        6.923, 7.692, 8.462, 9.231, 10.000, 11.538, 10.769, 12.308, 13.077,
        13.846, 14.615, 15.385, and 1.538 (again).
        Set with non-unity gain frames:: a set of subsets of frames. All frames
        in each subset have a unique, non-unity EM gain. For instance, in TVAC,
        11 subsets were considered with EM values (EMGAIN_C): 1.65, 5.24, 8.60,
        16.70, 27.50, 45.26, 87.50, 144.10, 237.26, 458.70 and 584.40. These
        correspond to a range of actual EM gains from about 2 to 7000. Each subset
        collects the same number of frames, which is at least 20 frames. In TVAC,
        each non-unity EM value had 22 frames. In each subset, there are two
        repeated exposure times: one near the start of the data collection and
        one at the very end. The exposure times of the frames in each EM subset
        do not need to be the same. For EM=1.65, the values of the exposure times
        in seconds were: 0.076, 0.758, 1.515, 2.273, 3.031, 3.789, 4.546, 5.304,
        6.062, 6.820, 7.577, 8.335, 9.093, 9.851, 10.608, 11.366, 12.124, 12.881,
        13.639, 14.397, 15.155, and 1.515 (repeated). And for EM=5.24, the 22
        values of the exposure times in seconds were: 0.070, 0.704, 1.408, 2.112,
        2.816, 3.520, 4.225, 4.929, 5.633, 6.337, 7.041, 7.745, 8.449, 9.153,
        9.857, 10.561, 11.265, 11.969, 12.674, 13.378, 14.082, and 1.408 (repeated).
      n_cal (int):
        Minimum number of sub-stacks used to calibrate Non-Linearity. The default
        value is 20.
      n_mean (int):
        Minimum number of frames used to generate the mean frame. The default value
        is 30.
      norm_val (int): (Optional) Value in DN to normalize the nonlinearity values to.
        Must be greater than 0 and must be divisible by 20 without remainder.
        (1500 to 3000 recommended).
      min_write (float): (Optional) Minimum mean value in DN to output in
        nonlin. (800.0 recommended)
      max_write (float): (Optional) Maximum mean value in DN to output in
        nonlin. (10000.0 recommended)
      lowess_frac (float): (Optional) factor to use in lowess smoothing function,
        larger is smoother
      rms_low_limit (float): (Optional) rms relative error selection limits for
        linear fit. Lower limit.
      rms_upp_limit (float): (Optional) rms relative error selection limits for
        linear fit. Upper limit. rms_upp_limit must be greater than rms_low_limit.
      pfit_upp_cutoff1 (int): (Optional) polyfit upper cutoff. The following limits were
        determined with simulated frames. If rms_low_limit < rms_y_rel_err < rms_upp_limit,
        this is the upper value applied to select the data to be fitted.
      pfit_upp_cutoff2 (int): (Optional) polyfit upper cutoff. The following limits were
        determined with simulated frames. If rms_y_rel_err >= rms_upp_limit,
        this is the upper value applied to select the data to be fitted.
      pfit_low_cutoff1 (int): (Optional) polyfit upper cutoff. The following limits were
        determined with simulated frames. If rms_low_limit < rms_y_rel_err < rms_upp_limit,
        this is the lower value applied to select the data to be fitted.
      pfit_low_cutoff2 (int): (Optional) polyfit upper cutoff. The following limits were
        determined with simulated frames. If rms_y_rel_err >= rms_upp_limit,
        this is the lower value applied to select the data to be fitted.
      make_plot (bool): (Optional) generate and store plots. Default is True.
      plot_outdir (str): (Optional) Output directory to store figues. Default is
        'figures'. The default directory is not tracked by git.
      show_plot (bool): (Optional) display the plots. Default is False.
      verbose (bool): (Optional) display various diagnostic print messages.
        Default is False.
      nonlin_params (dict): (Optional) Dictionary of row and col specifications
        for the region of interest (indicated by 'roi') where the frame is illuminated and for 
        two background regions (indicated by 'back1' and 'back2') where the frame is not illuminated.  
        Must contain 'rowroi1','rowroi2','colroi1','colroi2','rowback11','rowback12',
        'rowback21','rowback22','colback11','colback12','colback21',and 'colback22'.
        The 'roi' needs one square region specified, and 'back' needs two square regions, 
        where a '1' ending indicates the smaller of two values, and a '2' ending indicates the larger 
        of two values.  The coordinates of each square are specified by matching 
        up as follows: (rowroi1, colroi1), (rowroi1, colroi2), (rowback11, colback11), 
        (rowback11, colback12), etc. Defaults to nonlin_params_default specified in this file.
    
    Returns:
      nonlin_arr (NonLinearityCalibration): 2-D array with nonlinearity values
        for input signal level (DN) in rows and EM gain values in columns. The
        input signal in DN is the first column. Signal values start with min_write
        and run through max_write in steps of 20 DN.
    """
    if nonlin_params is None:
        nonlin_params = nonlin_params_default
        
    check_nonlin_params(nonlin_params)

    # dataset_nl.all_data must be 3-D 
    if np.ndim(dataset_nl.all_data) != 3:
        raise Exception('dataset_nl.all_data must be 3-D')
    # cast dataset objects into np arrays and retrieve aux information
    cal_arr, mean_frame_arr, exp_arr, datetime_arr, len_list, actual_gain_arr = \
        nonlin_dataset_2_stack(dataset_nl)
    # Get relevant constants
    rowroi1 = nonlin_params['rowroi1']
    rowroi2 = nonlin_params['rowroi2']
    colroi1 = nonlin_params['colroi1']
    colroi2 = nonlin_params['colroi2']
    rowback11 = nonlin_params['rowback11']
    rowback12 = nonlin_params['rowback12']
    rowback21 = nonlin_params['rowback21']
    rowback22 = nonlin_params['rowback22']
    colback11 = nonlin_params['colback11']
    colback12 = nonlin_params['colback12']
    colback21 = nonlin_params['colback21']
    colback22 = nonlin_params['colback22']
    min_exp_time = nonlin_params['min_exp_time']
    num_bins = nonlin_params['num_bins']
    min_bin = nonlin_params['min_bin']
    min_mask_factor = nonlin_params['min_mask_factor']

    if type(cal_arr) != np.ndarray:
        raise TypeError('cal_arr must be an ndarray.')
    if np.ndim(cal_arr) != 3:
        raise CalNonlinException('cal_arr must be 3-D')
    if len(len_list) < 1:
        raise CalNonlinException('Number of elements in len_list must '
                'be greater than or equal to 1.')
    if np.sum(len_list) != len(cal_arr):
        raise CalNonlinException('Number of sub-stacks in cal_arr must '
                'equal the sum of the elements in len_list')
    # cal_arr must have at least 20 frames for each EM gain
    if np.any(np.array(len_list) < n_cal):
        raise Exception(f'cal_arr must have at least {n_cal} frames for each EM value')
    if len(np.unique(datetime_arr)) != len(datetime_arr):
        raise CalNonlinException('All elements of datetime_arr must be unique.')
    for g_index in range(len(len_list)):
        # Define the start and stop indices
        start_index = int(np.sum(len_list[0:g_index]))
        stop_index = start_index + len_list[g_index]
        # Convert camera times to datetime objects
        ctim_strings = datetime_arr[start_index:stop_index]
        ctim_datetime = pd.to_datetime(ctim_strings, errors='coerce')
        # Check if the array is time-ordered in increasing order
        is_increasing = np.all(ctim_datetime[:-1] <= ctim_datetime[1:])
        if not is_increasing:
            raise CalNonlinException('Elements of datetime_arr must be '
                    'in increasing time order for each EM gain value.')
    if type(mean_frame_arr) != np.ndarray:
        raise TypeError('mean_frame_arr must be an ndarray.')
    if np.ndim(mean_frame_arr) != 3:
        raise CalNonlinException('mean_frame_arr must be 3-D (i.e., a stack of '
                '2-D sub-stacks')
    # mean_frame_arr must have at least 30 frames
    if len(mean_frame_arr) < n_mean:
        raise CalNonlinException(f'Number of frames in mean_frame_arr must '
                'be at least {n_mean}.')
    
    check.real_array(exp_arr, 'exp_arr', TypeError)
    check.oneD_array(exp_arr, 'exp_arr', TypeError)
    if (exp_arr <= min_exp_time).any():
        raise CalNonlinException('Each element of exp_arr must be '
            ' greater than min_exp_time.')
    index = 0
    r_flag = True
    for x in range(len(len_list)):
        temp = np.copy(exp_arr[index:index+len_list[x]])
        # Unique counts of exposure times
        _, u_counts = np.unique(temp, return_counts=True)
        # Check if all elements are the same
        all_elements_same = np.all(u_counts == u_counts[0])
        if all_elements_same == True:
            r_flag = False
        index = index + len_list[x]
    if not r_flag:
        raise CalNonlinException('each substack of cal_arr must have a '
            'group of frames with a repeated exposure time.')   
    if len(len_list) != len(actual_gain_arr):
        raise CalNonlinException('Length of actual_gain_arr be the same as the '
                                 'length of len_list.')
    if sum(1 for number in actual_gain_arr if number < 1) != 0:
        raise CalNonlinException('Each element of actual_gain_arr must be greater '
            'than or equal to 1.')
    check.real_array(actual_gain_arr, 'actual_gain_arr', TypeError)
    check.oneD_array(actual_gain_arr, 'actual_gain_arr', TypeError)
    check.positive_scalar_integer(norm_val, 'norm_val', TypeError)
    if np.mod(norm_val, 20) !=0:
        raise CalNonlinException('norm_val must be divisible by 20.')
    check.real_positive_scalar(min_write, 'min_write', TypeError)
    check.real_positive_scalar(max_write, 'max_write', TypeError)
    if min_write >= max_write:
        raise CalNonlinException('max_write must be greater than min_write')
    if (norm_val < min_write) or (norm_val > max_write):
        raise CalNonlinException('norm_val must be between min_write and '
                                 'max_write.')
    check.real_nonnegative_scalar(rms_low_limit, 'rms_low_limit', TypeError)
    check.real_nonnegative_scalar(rms_upp_limit, 'rms_upp_limit', TypeError)
    if rms_low_limit >= rms_upp_limit:
        raise CalNonlinException('rms_upp_limit must be greater than rms_low_limit')

    if not isinstance(lowess_frac, (float, int)):
        raise TypeError('lowess_frac is not a number')
    if not isinstance(rms_low_limit, (float, int)):
        raise TypeError('rms_low_limit is not a number')
    if not isinstance(rms_upp_limit, (float, int)):
        raise TypeError('rms_upp_limit is not a number')
    if not isinstance(pfit_upp_cutoff1, (float, int)):
        raise TypeError('pfit_upp_cutoff1 is not a number')
    if not isinstance(pfit_upp_cutoff2, (float, int)):
        raise TypeError('pfit_upp_cutoff2 is not a number')
    if not isinstance(pfit_low_cutoff1, (float, int)):
        raise TypeError('pfit_low_cutoff1 is not a number')
    if not isinstance(pfit_low_cutoff2, (float, int)):
        raise TypeError('pfit_low_cutoff2 is not a number')

    if make_plot is True:
        # Avoid issues with importing matplotlib on headless servers without GUI
        # support without proper configuration
        import matplotlib.pyplot as plt
        # Output directory
        if os.path.exists(plot_outdir) is False:
            os.mkdir(plot_outdir)
            if verbose:
                print('Output directory for figures created in ', os.getcwd())
    
    ######################### start of main code #############################
    
    # Define pixel ROIs
    rowroi = list(range(rowroi1, rowroi2))
    colroi = list(range(colroi1, colroi2))
    
    # Background subtraction regions
    rowback1 = list(range(rowback11, rowback12))
    rowback2 = list(range(rowback21, rowback22))
    colback1 = list(range(colback11, colback12))
    colback2 = list(range(colback21, colback22))
    
    ####################### create good_mean_frame ###################
    
    nrow = len(mean_frame_arr[0])
    ncol = len(mean_frame_arr[0][0])
    
    good_mean_frame = np.zeros((nrow, ncol))
    nFrames = len(mean_frame_arr)

    good_mean_frame = good_mean_frame / nFrames
    
    mean_frame_index = 0
    # Loop over the mean_frame_arr frames
    for i in range(nFrames):
        frame = mean_frame_arr[i]
    
        # Add this frame to the cumulative good_mean_frame
        good_mean_frame += frame
        mean_frame_index += 1

    # Calculate the average of the frames if required
    if mean_frame_index > 0:
        good_mean_frame /= mean_frame_index 
    
    # plot, if requested
    if make_plot:
        fname = 'non_lin_good_frame'
        # Slice the good_mean_frame array
        frame_slice = good_mean_frame[np.ix_(rowroi, colroi)]
        # Create a figure and plot the sliced frame
        plt.figure()
        # 'viridis' is a common colormap
        plt.imshow(frame_slice, aspect='equal', cmap='viridis')
        plt.colorbar()
        plt.title('Good quality mean frame')
        plt.savefig(f'{plot_outdir}/{fname}')
        if verbose:
            print(f'Figure {fname} stored in {plot_outdir}')
        if show_plot:
            plt.show()
        plt.close()
    
    # Convert to numpy arrays if they are not already
    rowroi = np.array(rowroi)
    colroi = np.array(colroi)
    
    if make_plot:
        fname = 'non_lin_mean_frame_histogram'
        # Plot a histogram of the values within the specified ROI
        roi_values = good_mean_frame[rowroi[:, None], colroi]
        plt.figure()
        # 'auto' lets matplotlib decide the number of bins
        plt.hist(roi_values.flatten(), bins='auto', log=True)
        plt.gca().set_yscale('log')
        plt.gca().set_xscale('log')
        plt.title('Histogram of Mean Frame in ROI')
        plt.savefig(f'{plot_outdir}/{fname}')
        if verbose:
            print(f'Figure {fname} stored in {plot_outdir}')
        if show_plot:
            plt.show()
        plt.close()
    
    # find minimum in histogram
    # 1000-1500 DN recommended when the peak of histogram of  
    # "good_mean_frame" is between 2000 and 4000 DN)
    roi_values = good_mean_frame[rowroi[:, None], colroi]
    hst_counts, hist_edges = np.histogram(roi_values.flatten(),bins=num_bins)
    # range above some value
    above_range = (hist_edges[:-1] >= min_bin)
    # Filter the counts and bin_edges arrays
    filtered_counts_above = hst_counts[above_range]
    filtered_bin_edges_above = hist_edges[:-1][above_range]
    # Find the index of the maximum count within the filtered range
    max_count_index_above_range = np.argmax(filtered_counts_above)
    # Get the corresponding bin edge
    max_edge_value = filtered_bin_edges_above[max_count_index_above_range]
    # Find the indices of the bins that fall within the specified range
    within_range = (hist_edges[:-1] >= min_bin) & (hist_edges[:-1] <= max_edge_value)
    # Filter the counts and bin_edges arrays
    filtered_counts = hst_counts[within_range]
    filtered_bin_edges = hist_edges[:-1][within_range]
    # Find the index of the minimum count within the filtered range
    min_count_index_within_range = np.argmin(filtered_counts)
    # Get the corresponding bin edge value and increase by min_mask_factor
    min_mask = min_mask_factor*filtered_bin_edges[min_count_index_within_range]
    # Create the mask
    mask = np.where(good_mean_frame < min_mask, 0, 1)
    
    # plot, if requested
    if make_plot:
        fname = 'non_lin_mask'
        # Plot the mask
        plt.figure()
        plt.imshow(mask, cmap='gray')
        plt.title('Mask')
        plt.colorbar()
        plt.savefig(f'{plot_outdir}/{fname}')
        if verbose:
            print(f'Figure {fname} stored in {plot_outdir}')
        if show_plot:
            plt.show()
        plt.close()
        
        fname = 'non_lin_mean_frame'
        # Plot the mean frame
        plt.figure()
        # 'viridis' is a good default color map
        plt.imshow(good_mean_frame, cmap='viridis')
        plt.title('Mean Frame')
        plt.colorbar()
        plt.close()
    
    # initialize arrays for nonlin results table
    nonlin = []
    
    ######################## loop over em gain values #########################
    for gain_index in range(len(len_list)):
        
        start_index = int(np.sum(len_list[0:gain_index]))
        stop_index = start_index + len_list[gain_index]
        # Convert camera times to datetime objects
        ctime_strings = datetime_arr[start_index:stop_index]
        ctime_datetime = pd.to_datetime(ctime_strings, errors='coerce')
        
        # Select exp times for this em gain
        exp_em = exp_arr[start_index:stop_index]
        
        # select frames for this em gain
        full_flst = cal_arr[start_index:stop_index]
        
        # Unique exposure times and their counts
        exposure_strings_list, counts = np.unique(exp_em, return_counts=True)
        
        # Grouping exposures and finding the max count
        max_count_index = np.argmax(counts)
        repeat_exp = exposure_strings_list[max_count_index]  # Exposure time of repeated frames
        
        # Calculate mean time differences as aid in illumination drift corrections
        group_mean_time = []
        first_flag = False
        
        for t0 in exposure_strings_list:
            idx = np.where(exp_em == t0)[0]
            if t0 != repeat_exp:
                del_s = (ctime_datetime[idx] - ctime_datetime[0]).total_seconds()
                group_mean_time.append(np.mean(del_s))
            elif t0 == repeat_exp and not first_flag:
                idx_2 = len(idx) // 2
                del_s = (ctime_datetime[idx[:idx_2]] - ctime_datetime[0]).total_seconds()
                group_mean_time.append(np.mean(del_s))
                first_flag = True
        
        if verbose is True:
            print(group_mean_time)
        
        # Additional setup
        mean_signal = []
        repeat_flag = 0
        filtered_exposure_times = []
        
        for jj in range(len(exposure_strings_list)):
            current_exposure_time = exposure_strings_list[jj]
        
            if current_exposure_time >= min_exp_time:
                if current_exposure_time == repeat_exp:
                    repeat_flag = 1
        
                # Filtering frames based on the current exposure time
                selected_files = [
                    full_flst[idx] for idx, exp_time in enumerate(exp_em) if exp_time == current_exposure_time
                ]

                filtered_exposure_times.append(current_exposure_time)
        
                # Initialize for processing of files
                mean_frame_index = 0
                frame_count = []
                frame_mean = []
                if not repeat_flag:
                    for iframe in range(len(selected_files)):
                        
                        frame_1 = selected_files[iframe]
                        frame_1 = frame_1.astype(np.float64)
        
                        # Subtract background
                        frame_1_back1 = np.mean(frame_1[rowback1[0]:rowback1[-1]+1, 
                                                        colback1[0]:colback1[-1]+1])
                        frame_1_back2 = np.mean(frame_1[rowback2[0]:rowback2[-1]+1, 
                                                        colback2[0]:colback2[-1]+1])
                        frame_back = (frame_1_back1 + frame_1_back2) / 2
        
                        # Calculate counts and mean in the ROI after background subtraction
                        roi_frame = frame_1[rowroi[0]:rowroi[-1]+1, colroi[0]:colroi[-1]+1] - frame_back
                        frame_count0 = np.sum(roi_frame)
                        frame_mean0 = frame_1 - frame_back
        
                        # Apply mask and calculate the positive mean
                        frame_mean0 *= mask
                        positive_means = frame_mean0[frame_mean0 > 0]
                        frame_mean1 = np.mean(positive_means) if positive_means.size > 0 else np.nan
        
                        frame_count.append(frame_count0)
                        frame_mean.append(frame_mean1)
                        
                        mean_frame_index += 1
                    mean_signal.append(np.mean(frame_mean))
                elif repeat_flag:
                    # for repeated exposure frames, split into the first half/set
                    # and the second half/set
                    first_half = len(selected_files) // 2
                    for i in range(first_half):

                        frame_1 = selected_files[i]
                        frame_1 = frame_1.astype(np.float64)
        
                        # Subtract background
                        frame_1_back1 = np.mean(frame_1[rowback1[0]:rowback1[-1]+1, 
                                                        colback1[0]:colback1[-1]+1])
                        frame_1_back2 = np.mean(frame_1[rowback2[0]:rowback2[-1]+1, 
                                                        colback2[0]:colback2[-1]+1])
                        frame_back = (frame_1_back1 + frame_1_back2) / 2
        
                        # Calculate counts and mean in the ROI after background subtraction
                        roi_frame = frame_1[rowroi[0]:rowroi[-1]+1, 
                                            colroi[0]:colroi[-1]+1] - frame_back
                        frame_count0 = np.sum(roi_frame)
                        frame_mean0 = frame_1 - frame_back
        
                        # Apply mask and calculate the positive mean
                        frame_mean0 *= mask
                        positive_means = frame_mean0[frame_mean0 > 0]
                        frame_mean1 = np.mean(positive_means) if positive_means.size > 0 else np.nan
                        
                        frame_count.append(frame_count0)
                        frame_mean.append(frame_mean1)
                        
                        mean_frame_index += 1
                    mean_signal.append(np.nanmean(frame_mean))
                    repeat1_mean_signal = np.nanmean(frame_mean)
                    
                    second_half = len(selected_files)
                    for i in range(first_half + 1, second_half):
                       
                        frame_1 = selected_files[i]
                        frame_1 = frame_1.astype(np.float64)
        
                        # Subtract background
                        frame_1_back1 = np.mean(frame_1[rowback1[0]:rowback1[-1]+1, colback1[0]:colback1[-1]+1])
                        frame_1_back2 = np.mean(frame_1[rowback2[0]:rowback2[-1]+1, colback2[0]:colback2[-1]+1])
                        frame_back = (frame_1_back1 + frame_1_back2) / 2
        
                        # Calculate counts and mean
                        roi_frame = frame_1[rowroi[0]:rowroi[-1]+1, colroi[0]:colroi[-1]+1] - frame_back
                        frame_count0 = np.sum(roi_frame)
                        frame_mean0 = frame_1 - frame_back
                        frame_mean0 *= mask
                        positive_means = frame_mean0[frame_mean0 > 0]
                        frame_mean1 = np.mean(positive_means) if positive_means.size > 0 else np.nan
        
                        frame_count.append(frame_count0)
                        frame_mean.append(frame_mean1)
        
                        mean_frame_index += 1
                    # Calculate the mean signal from the second half of the processing
                    repeat2_mean_signal = np.nanmean(frame_mean)
                    repeat_flag = 0  # Reset flag

        # Calculate the time deltas in seconds from the first frame
        delta_ctimes_s = (ctime_datetime - ctime_datetime[0]).total_seconds()
        
        # Make sure delta_ctimes_s is a pandas Series with numeric values
        delta_ctimes_s = pd.Series(delta_ctimes_s, index=ctime_datetime)
        
        # Calculate the difference in signals
        delta_signal = repeat2_mean_signal - repeat1_mean_signal
        
        # Assuming all_exposure_strings and repeat_exp are already defined
        
        # Find indices of the frames where the exposure time matches repeat_exp
        repeat_times_idx = np.where(exp_em == repeat_exp)[0]  # np.where returns a tuple, extract first element
        
        # Calculate the mean times for the first and second halves of these indices
        first_half = len(repeat_times_idx) // 2
        first_half_mean_time = delta_ctimes_s.iloc[repeat_times_idx[:first_half]].mean()
        
        second_half = len(repeat_times_idx)
        second_half_mean_time = delta_ctimes_s.iloc[repeat_times_idx[first_half:second_half]].mean()
        
        if verbose is True:
            print("First half mean time:", first_half_mean_time)
            print("Second half mean time:", second_half_mean_time)
        
        # Calculate DN/s
        illum_slope = delta_signal / (second_half_mean_time - first_half_mean_time)
        
        # Calculate DN
        illum_inter = repeat1_mean_signal - illum_slope * first_half_mean_time
        
        # Adjust observations based on calculated slope and intercept
        illum_obs = (group_mean_time - group_mean_time[0]) * illum_slope + illum_inter
        
        # Correct the illumination observations
        illum_corr = illum_obs / illum_obs[0]
        
        # Correct the mean signal
        #illum_cor = np.ones(len(illum_corr))
        corr_mean_signal = mean_signal / illum_corr
        
        # Sort arrays by exposure time
        filt_exp_times_sorted, I = np.sort(filtered_exposure_times), np.argsort(filtered_exposure_times)
        corr_mean_signal_sorted = np.array(corr_mean_signal)[I]
        
        if make_plot:
            fname = 'non_lin_signal_vs_exp'
            # Plotting the corrected mean signal against sorted exposure times
            plt.figure()
            plt.plot(filt_exp_times_sorted, corr_mean_signal_sorted, 'o', label='Data Points')
            plt.title('Signal versus exposure time')
            plt.xlabel('Exposure time (s)')
            plt.ylabel('Signal (DN)')
        
        # Fit a polynomial to selected points (excluding some points)
        p0 = np.polyfit(filt_exp_times_sorted, corr_mean_signal_sorted, 1)
        y0 = np.polyval(p0, filt_exp_times_sorted)
        y_rel_err = np.abs((corr_mean_signal_sorted - y0)/corr_mean_signal_sorted)
        rms_y_rel_err = np.sqrt(np.mean(y_rel_err**2))
        # NOTE: the following limits were determined with simulated frames
        if rms_y_rel_err < rms_low_limit:
            p1 = np.polyfit(filt_exp_times_sorted, corr_mean_signal_sorted, 1)
        elif (rms_y_rel_err >= rms_low_limit) and (rms_y_rel_err < rms_upp_limit):
            p1 = np.polyfit(filt_exp_times_sorted[pfit_low_cutoff1:pfit_upp_cutoff1], 
                            corr_mean_signal_sorted[pfit_low_cutoff1:pfit_upp_cutoff1], 1)
        else:
            p1 = np.polyfit(filt_exp_times_sorted[pfit_low_cutoff2:pfit_upp_cutoff2], 
                            corr_mean_signal_sorted[pfit_low_cutoff2:pfit_upp_cutoff2], 1)
        y1 = np.polyval(p1, filt_exp_times_sorted)
        
        if make_plot:
            fname = 'non_lin_fit'
            # Plot the fitted line
            plt.plot(filt_exp_times_sorted, y1, label='Fitted Line')
            
            # Show the plot with legend
            plt.legend()
            plt.savefig(f'{plot_outdir}/{fname}')
            if verbose:
                print(f'Figure {fname} stored in {plot_outdir}')
            if show_plot:
                plt.show()
            plt.close()
        
        # Calculating relative gain
        rel_gain = corr_mean_signal_sorted / y1
        
        # Smoothing the relative gain data; larger 'lowess_frac' gives smoother curve
        rel_gain_smoothed = lowess(rel_gain, 
                            corr_mean_signal_sorted, frac=lowess_frac)[:, 1]
        
        # find the min/max values of corrected measured means and append array
        temp_min = np.min(corr_mean_signal_sorted)
        temp_max = np.max(corr_mean_signal_sorted)
        
        if make_plot:
            # Plotting Signal vs. Relative Gain
            plt.figure()
            plt.plot(corr_mean_signal_sorted, rel_gain, 'o', label='Original Data')
            plt.ylim([0.95, 1.05])
            plt.xlim([1, 14000])
            plt.axhline(1.0, linestyle='--', color='k', linewidth=1)  # horizontal line at 1.0
            
            plt.title('Signal/fit versus Signal')
            plt.xlabel('Signal (DN)')
            plt.ylabel('Relative gain')
            
            # Plot the smoothed data
            plt.plot(corr_mean_signal_sorted, rel_gain_smoothed, 'r-', label='Smoothed Data')
            
            # Show legend and plot
            plt.legend()
            plt.savefig(f'{plot_outdir}/{fname}')
            if verbose:
                print(f'Figure {fname} stored in {plot_outdir}')
            if show_plot:
                plt.show()
            plt.close()
        
        # Generate evenly spaced values between 20 and 14000
        mean_linspace = np.linspace(20, 14000, 1+int((14000-20)/20))
        
        # Interpolate/extrapolate the relative gain values
        interp_func = interp1d(corr_mean_signal_sorted, 
                        rel_gain_smoothed, kind='linear', fill_value='extrapolate')
        rel_gain_interp = interp_func(mean_linspace)
        
        # Normalize the relative gain to the value at norm_val DN
        # First, find the index for norm_val DN in mean_linspace
        idxnorm = np.where(mean_linspace == norm_val)[0][0]
        normconst = rel_gain_interp[idxnorm]
        rel_gain_interp /= normconst
        if (norm_val < temp_min) or (norm_val > temp_max):
            warnings.warn('norm_val is not between the minimum and maximum values '
                          'of the means for the current EM gain. Extrapolation '
                          'will be used for norm_val.')
        
        if make_plot:
            fname = 'non_lin_fit_norm_dn'
            # Plotting Signal vs. Relative Gain normalized at norm_val DN
            plt.figure()
            plt.plot(corr_mean_signal_sorted, rel_gain / normconst, 'o', label='Original Data')
            plt.ylim([0.95, 1.05])
            plt.xlim([1, 14000])
            plt.axhline(1.0, linestyle='--', color='k', linewidth=1)  # horizontal line at 1.0
            
            plt.title(f'Signal/fit versus Signal (norm @ {norm_val} DN)')
            plt.xlabel('Signal (DN)')
            plt.ylabel('Relative gain')
            
            # Plot the interpolated data
            plt.plot(mean_linspace, rel_gain_interp, 'r-', label='Interpolated Data')
            plt.legend()
            plt.savefig(f'{plot_outdir}/{fname}')
            if verbose:
                print(f'Figure {fname} stored in {plot_outdir}')
            if show_plot:
                plt.show()
            plt.close()
        
        # NOTE: nonlinearity is equal to 1/rel_gain
        # multiply raw data by 1/rel_gain to correct for nonlinearity
        temp = 1/rel_gain_interp
        nonlin.append(temp)
    
    # prepare nonlin array
    nonlin_arr0 = np.transpose(np.array(nonlin))
    # insert new column at the start of nonlin_arr
    nonlin_arr1 = np.insert(nonlin_arr0, 0, mean_linspace, axis=1)
    # select rows that satisfy min/max limits
    nonlin_arr2 = nonlin_arr1[nonlin_arr1[:, 0] >= min_write]
    nonlin_arr3 = nonlin_arr2[nonlin_arr2[:, 0] <= max_write]
    # See data.NonLinearityCalibration doc string for more details:
    # [0, 1:]: Gain axis values
    # [1:, 0]: "count" axis value
    actual_gain_arr = np.insert(actual_gain_arr, 0, np.nan)
    n_col = len(nonlin_arr3) + 1
    n_row = len(actual_gain_arr)
    nonlin_data=np.insert(nonlin_arr3, 0, actual_gain_arr).reshape(n_col,n_row)
    
    # Return NonLinearity instance
    prhd = dataset_nl.frames[0].pri_hdr
    exthd = dataset_nl.frames[0].ext_hdr
    exthd['HISTORY'] = f"Non-linearity calibration derived from a set of frames on {exthd['DATETIME']}"
    # Just for the purpose of getting the instance created
    nonlin = data.NonLinearityCalibration(nonlin_data,
        pri_hdr = prhd, ext_hdr = exthd, input_dataset=dataset_nl)
    
    return nonlin

def nonlin_dataset_2_stack(dataset):
    """
    Casts the CORGIDRP Dataset object for non-linearity calibration into a stack
    of numpy arrays sharing the same commanded gain value. It also returns the list of
    unique EM values and set of exposure times used with each EM. Note: it also
    performs a set of tests about the integrity of the data type and values in
    the dataset.

    Args:
        dataset (corgidrp.Dataset): Dataset with a set of of EXCAM illuminated
        pupil L1 SCI frames (counts in DN)

    Returns:
        numpy array with stack of stacks of data array associated with each frame
        array of exposure times associated with each frame
        array of datetimes associated with each frame
        list with the number of frames with same EM gain
        List of actual EM gains

    """
    # Split Dataset
    dataset_cp = dataset.copy()
    split = dataset_cp.split_dataset(exthdr_keywords=['EMGAIN_C'])
    
    # Calibration data
    stack = []
    # Mean frame data
    mean_frame_stack = []
    record_exp_time = True
    # Exposure times
    exp_times = []
    # Datetimes
    datetimes = []
    # Size of each sub stack
    len_sstack = []
    # Record measured gain of each substack of calibration frames
    gains = []
    for idx_set, data_set in enumerate(split[0]):
        # Second layer for calibration data (array of different exposure times)
        sub_stack = []
        len_cal_frames = 0
        record_gain = True 
        for frame in data_set.frames:
            if frame.pri_hdr['OBSNAME'] == 'MNFRAME':
                if record_exp_time:
                    exp_time_mean_frame = frame.ext_hdr['EXPTIME'] 
                    record_exp_time = False
                if frame.ext_hdr['EXPTIME'] != exp_time_mean_frame:
                    raise Exception('Frames used to build the mean frame must have the same exposure time')
                if frame.ext_hdr['EMGAIN_C'] != 1:
                    raise Exception('The commanded gain used to build the mean frame must be unity')
                mean_frame_stack.append(frame.data)
            elif (frame.pri_hdr['OBSNAME'] == 'KGAIN' or
                frame.pri_hdr['OBSNAME'] == 'NONLIN'):
                len_cal_frames += 1
                sub_stack.append(frame.data)
                exp_time = frame.ext_hdr['EXPTIME']
                if isinstance(exp_time, float) is False:
                    raise Exception('Exposure times must be float')
                if exp_time <=0:
                    raise Exception('Exposure times must be positive')
                exp_times.append(exp_time)
                datetime = frame.ext_hdr['DATETIME']
                
                if isinstance(datetime, str) is False:
                    raise Exception('DATETIME must be a string')
                datetimes.append(datetime)
                if record_gain:
                    try: # if EM gain measured directly from frame TODO change hdr name if necessary
                        gains.append(frame.ext_hdr['EMGAIN_M'])
                    except:
                        if frame.ext_hdr['EMGAIN_A'] > 0: # use applied EM gain if available
                            gains.append(frame.ext_hdr['EMGAIN_A'])
                        else: # use commanded gain otherwise
                            gains.append(frame.ext_hdr['EMGAIN_C'])
                        record_gain = False
            else:
                raise Exception('OBSNAME can only be MNFRAME or NONLIN in non-linearity')
        
        # First layer (array of unique EM values)
        if len(sub_stack):
            stack.append(np.stack(sub_stack))
            len_sstack.append(len_cal_frames)

    # All elements of datetimes must be unique
    if len(datetimes) != len(set(datetimes)):
        raise Exception('DATETIMEs cannot be duplicated')
    # Length of substack must be at least 1
    if len(len_sstack) == 0:
        raise Exception('Substacks must have at least one element')
    # Every EM gain must be greater than or equal to 1
    if np.any(np.array(split[1]) < 1):
        raise Exception('Each set of frames categorized by commanded EM gains must be have 1 or more frames')
    if np.any(np.array(gains) < 1):
        raise Exception('Actual EM gains must be greater than or equal to 1')

    return (np.vstack(stack), np.stack(mean_frame_stack), np.array(exp_times),
        np.array(datetimes), len_sstack, np.array(gains))



import os
import numpy as np
from astropy.time import Time
from astropy.io import fits, ascii

import corgidrp
from corgidrp import astrom, data

here = os.path.abspath(os.path.dirname(__file__))

def get_cfam(
    cfam_name='1F',
    cfam_version=0,
    ):
    """ Read CFAM filter wavelength in nm and transmission.

    Args:
      cfam_name (string): Filter in CFAM. For instance, '1F', '4A', '3B' or '2C'.
      cfam_version (int): version number of the filters (CFAM, pupil, imaging
        lens).

    Returns:
      CFAM filter wavelength in nm and transmission.
    """
    datadir = os.path.join(here, 'data', 'filter_curves')
    filter_names = os.listdir(datadir)
    filter_name = [name for name in filter_names if name.find(cfam_name) >= 0]
    if filter_name == []:
        raise ValueError(f'there is no filter available with name {cfam_name}')
    filter_name = [name for name in filter_name if f'v{cfam_version}' in name]
    if filter_name == []:
        raise ValueError(f'there is no filter {cfam_name} available with version {cfam_version}')
    tab = ascii.read(os.path.join(datadir,filter_name[0]), format='csv',
        header_start = 3, data_start = 4)
    lambda_nm_filter = tab['lambda_nm'].data
    trans_filter = tab['%T'].data / tab['%T'].data.max()
    return lambda_nm_filter, trans_filter

def di_over_pil_transmission(
    cfam_name='1F',
    cfam_version=0,
    ):
    """ Derives the relative transmission between the pupil lens and the imaging
      lens: trans_imaging/trans_pupil.
 
      Multiplying the counts of the pupil image by this factor translates them
      into equivalent counts of the direct imaging lens.
 
    Args:
      cfam_name (string): Filter in CFAM. For instance, '1F', '4A', '3B' or '2C'.
      cfam_version (int): version number of the filters (CFAM, pupil, imaging
        lens).

    Returns:
      Ratio trans_imaging/trans_pupil.
    """
    # Read pupil and direct imaging lenses
    try:
        lambda_pupil_A, trans_pupil = np.loadtxt(os.path.join(here, 'data',
            'filter_curves', f'pupil_lens_v{cfam_version}.txt'),
            delimiter=',', unpack=True)
        lambda_pupil_nm = lambda_pupil_A / 10
    except:
        raise Exception('* File with the transmission of the pupil lens not found')

    try:
        lambda_imaging_A, trans_imaging = np.loadtxt(os.path.join(here, 'data',
            'filter_curves', f'imaging_lens_v{cfam_version}.txt'),
            delimiter=',', unpack=True)
        lambda_imaging_nm = lambda_imaging_A / 10
    except:
        raise Exception('* File with the transmission of the imaging lens not found')

    # Get CFAM filter wavelength and transmission
    lambda_nm_filter, trans_lambda_filter = get_cfam(cfam_name=cfam_name,
        cfam_version=cfam_version)

    # Linear interpolation
    trans_lambda_pupil_band = np.interp(
        lambda_nm_filter,
        lambda_pupil_nm,
        trans_pupil)
    trans_lambda_imaging_band = np.interp(
        lambda_nm_filter,
        lambda_imaging_nm,
        trans_imaging)
    # Ratio of both transmissions:
    ratio_imaging_pupil_trans = (np.sum(trans_lambda_imaging_band*trans_lambda_filter)/
        np.sum(trans_lambda_pupil_band*trans_lambda_filter))
    return ratio_imaging_pupil_trans

def get_psf_pix(
    dataset,
    roi_radius=3,
    ):
    """ Estimate the PSF positions of a set of PSF images. 
 
    Args:
      dataset (corgidrp.data.Dataset): a collection of off-axis PSFs.
      roi_radius (int or float): Half-size of the box around the peak,
        in pixels. Adjust based on desired λ/D.

    Returns:
      Array of pair of values with PSFs position in (fractional) EXCAM pixels
      with respect to the pixel (0,0) in the PSF images
    """
    psf_pix = []
    for psf in dataset:
        psf_pix += [astrom.centroid_with_roi(psf.data,roi_radius=roi_radius)]
    return np.array(psf_pix)

def get_psf_ct(
    dataset,
    unocc_psf_norm=1,
    ):
    """ Estimate the core throughput of a set of PSF images.

    Definition of core throughput: The numerator in CT (counts above 50% peak)
    is measured with pupil masks (Lyot stop, SPC pupil mask) in place, DMs at
    dark hole solution, but no FPM.  The denominator (total stellar flux) is
    measured without any masks in place and an infinite aperture.

    NOTE: The FPM are kept in place while measuring the CT because near the
    region of 6 lam/D, the FPM effects are negligible and the CT data set allows
    one to quantify the effect of the FPM in other areas, near the IWA and OWA,
    respectively.

    See  Journal of Astronomical Telescopes, Instruments, and Systems, Vol. 9,
    Issue 4, 045002 (October 2023). https://doi.org/10.1117/1.JATIS.9.4.045002
    and figures 9-13 for details.

    Args:
      dataset (corgidrp.data.Dataset): a collection of off-axis PSFs.
      unocc_psf_norm (float): sum of the 2-d array corresponding to the
        unocculted psf. Default: off-axis PSF are normalized to the unocculted
        PSF already. That is, unocc_psf_norm equals 1.

    Returns:
      Array of core throughput values between 0 and 1.
    """
    psf_ct = []
    for psf in dataset:
        psf_ct += [psf.data[psf.data >= psf.data.max()/2].sum()/unocc_psf_norm]
    psf_ct = np.array(psf_ct, dtype=float)

    return psf_ct

def estimate_psf_pix_and_ct(
    dataset_in,
    roi_radius=3,
    cfam_version=0,
    ):
    """
    1090881 - Given a core throughput dataset consisting of M clean frames
    (nominally 1024x1024) taken at different FSM positions, the CTC GSW shall
    estimate the pixel location and core throughput of each PSF.

    NOTE: the list of M clean frames may be a subset of the frames collected during
    core throughput data collection, to allow for the removal of outliers.

    Some of the images are pupil images of the unocculted source.

    Args:
      dataset_in (corgidrp.data.Dataset): A core throughput dataset consisting of
        M clean frames (nominally 1024x1024) taken at different FSM positions.
        It includes some pupil images of the unocculted source.  photoelectrons / second / pixel.
      roi_radius (int or float): Half-size of the box around the peak,
        in pixels. Adjust based on desired λ/D.
      cfam_version (int): version number of the filters (CFAM, pupil, imaging
        lens).

    Returns:
      psf_pix (array): Array with PSF's pixel positions. Units: EXCAM pixels
        referred to the (0,0) pixel.
      psf_ct (array): Array with PSF's core throughput values. Units:
        dimensionless (Values must be within 0 and 1).
    """
    dataset = dataset_in.copy()

    # All frames must have the same CFAM filter
    cfam_list = []
    for frame in dataset:
        try:
            cfam_list += [frame.ext_hdr['CFAMNAME']]
        except:
            raise Exception('Frame w/o CFAM specification. All frames must have CFAM specified')
    if len(set(cfam_list)) != 1:
        raise Exception('All frames must have the same CFAM filter')
    
    # identify the pupil images in the dataset
    pupil_img_frames = []
    for frame in dataset:
        try:
        # Pupil images of the unocculted source satisfy:
        # DPAM=PUPIL, LSAM=OPEN, FSAM=OPEN and FPAM=OPEN_12
            exthd = frame.ext_hdr
            if (exthd['DPAMNAME']=='PUPIL' and exthd['LSAMNAME']=='OPEN' and
                exthd['FSAMNAME']=='OPEN' and exthd['FPAMNAME']=='OPEN_12'):
                pupil_img_frames += [frame]
        except:
            pass
    if pupil_img_frames:
        print(f'Found {len(pupil_img_frames)} pupil images for the core throughput estimation.')
    else:
        raise Exception('No pupil image found. At least there must be one pupil image.')
    # mean combine the total values (photo-electrons/sec) of the pupil images
    unocc_psf_norm = 0
    for frame in pupil_img_frames:
        unocc_psf_norm += frame.data.sum()
    unocc_psf_norm /= len(pupil_img_frames)
    # Transform pupil counts into direct imaging counts. Recall all frames have
    # the same cfam filter or an Exception is raised
    unocc_psf_norm *= di_over_pil_transmission(cfam_name=cfam_list[0],
        cfam_version=cfam_version)
    # Remove pupil frames
    offaxis_frames = []
    for frame in dataset:
        if frame not in pupil_img_frames:
            offaxis_frames += [frame]
    dataset_offaxis = corgidrp.data.Dataset(offaxis_frames)
    if len(dataset_offaxis):
        print(f'Found {len(dataset_offaxis)} off-axis PSFs for the core throughput estimation.')
    else:
        raise Exception('No off-axis PSF found. At least there must be one off-axis PSF.')
    # find the PSF positions of the off-axis PSFs
    psf_pix = get_psf_pix(
        dataset_offaxis,
        roi_radius=roi_radius)
    # find the PSF corethroughput of the off-axis PSFs
    psf_ct = get_psf_ct(
        dataset_offaxis,
        unocc_psf_norm = unocc_psf_norm)
    # same number of estimates. One per PSF
    if len(psf_pix) != len(psf_ct) or len(psf_pix) != len(dataset_offaxis):
        raise Exception('PSF positions and CT values are inconsistent')
    return psf_pix, psf_ct

def generate_psf_cube(
    dataset_in,
    psf_loc,
    cfam_name='1F',
    cfam_version=0,
    ):
    """
    Function that derives a 3-d cube of PSF images from a core throughput dataset.

    # TODO: error data cubes will be added in a release after R3.0.2

    Args:
      dataset_in (corgidrp.data.Dataset): A core throughput dataset consisting of
        M clean frames (nominally 1024x1024) taken at different FSM positions.
        It includes some pupil images of the unocculted source.
      psf_loc (array): Array of pair of values with PSFs position in (fractional)
        EXCAM pixels with respect to the pixel (0,0) in the PSF images.
      cfam_name (string): Filter in CFAM. For instance, '1F', '4A', '3B' or '2C'.
      cfam_version (int): version number of the filters (CFAM, pupil, imaging
        lens).

    Returns:
      3-d PSF cube of PSF images from a core throughput dataset, including their
      data quality, and corresponding headers as HDU units. 
    """
    dataset = dataset_in.copy()

    # 3-d cube of PSF images cut around the PSF's location
    psf_cube = []
    dq_cube = []
    # Pixels arounf PSF's location +/- n_pix_psf in both dimensions that
    # correspond to 3 lambda/D in units of EXCAM pixels:
    # 3 * lambda_mean_nm * 1e-9 / D * rad_to_mas / EXCAM_pixel_pitch in mas
    n_pix_psf = int(np.ceil(3*get_cfam(cfam_name=cfam_name,
        cfam_version=cfam_version)[0].mean()*1e-9/2.36*180/np.pi*3600e3/21.8))
    i_psf = 0
    for frame in dataset:
        # Skip pupil images of the unocculted source, which satisfy:
        # DPAM=PUPIL, LSAM=OPEN, FSAM=OPEN and FPAM=OPEN_12
        try:
        # Pupil images of the unocculted source satisfy:
        # DPAM=PUPIL, LSAM=OPEN, FSAM=OPEN and FPAM=OPEN_12
            exthd = frame.ext_hdr
            if (exthd['DPAMNAME']=='PUPIL' and exthd['LSAMNAME']=='OPEN' and
                exthd['FSAMNAME']=='OPEN' and exthd['FPAMNAME']=='OPEN_12'):
                continue
        except:
            pass
        idx_0_0 = max(int(np.round(psf_loc[i_psf][1])) - n_pix_psf,0)
        idx_0_1 = min(frame.data.shape[0],
            int(np.round(psf_loc[i_psf][1])) + n_pix_psf + 1)
        idx_1_0 = max(int(np.round(psf_loc[i_psf][0])) - n_pix_psf,0)
        idx_1_1 = min(frame.data.shape[1],
            int(np.round(psf_loc[i_psf][0])) + n_pix_psf + 1)
        psf_cube += [frame.data[idx_0_0:idx_0_1, idx_1_0:idx_1_1]]
        dq_cube += [frame.dq[idx_0_0:idx_0_1, idx_1_0:idx_1_1]]
        i_psf += 1

    psf_cube = np.array(psf_cube)
    dq_cube = np.array(dq_cube)
    # Check
    if len(psf_cube) != len(psf_loc):
        raise Exception(('The number of PSFs does not match the number of PSF '+
            ' locations.'))

    # PSF cube header
    ext_hdr = dataset[0].ext_hdr
    # Add history
    ext_hdr['HISTORY'] = ('Core Throughput calibration derived from a '
        f'set of frames from {dataset[0].ext_hdr["DATETIME"]} to '
        f'{dataset[-1].ext_hdr["DATETIME"]}')
    # Add specific information
    ext_hdr['UNITS'] = 'photoelectron/pix/s'
    ext_hdr['COMMENT'] = ('Set of PSFs derived from a core throughput '
        'observing sequence. PSFs are not normalized. They are the images of the '
        'off-axis source. The data cube is centered around each PSF location')
    # Add EXTNAME
    psf_hdu = fits.ImageHDU(data=psf_cube, header=ext_hdr, name='PSFCUBE')

    # Data quality cube
    dq_hdr = dataset[0].dq_hdr
    # Add specific information
    dq_hdr['COMMENT'] = 'Data quality for each image' 
    # Add EXTNAME
    dq_hdu = fits.ImageHDU(data=dq_cube, header=dq_hdr, name='DQCUBE')

    return psf_hdu, dq_hdu

def generate_ct_cal(
    dataset_in,
    roi_radius=3,
    cfam_version=0,
    ):
    """
    Generate the elements needed to create a core throughput calibration file.

    A CoreThroughput calibration file has two main data arrays:

    3-d cube of PSF images, e.g, a N1xN1xN array where N1= +/- 3l/D about 
      PSF's centroid in EXCAM pixels. The N PSF images are the ones in the CT
      dataset.

    N sets of (x, y, CT measurements). The (x, y) are pixel coordinates of the
      PSF images in the CT dataset wrt EXCAM (0,0) pixel during core throughput
      observation.

    Args:
      dataset_in (corgidrp.data.Dataset): A core throughput dataset consisting of
        M clean frames (nominally 1024x1024) taken at different FSM positions.
        It includes some pupil images of the unocculted source.
      roi_radius (int or float): Half-size of the box around the peak,
        in pixels. Adjust based on desired λ/D.
      cfam_version (int): version number of the filters (CFAM, pupil, imaging
        lens).

    Returns:
      PSF cube, data quality cube, HDU list with the CT array measurements,
      including the PSF locations, FPAM/FSAM positions, and corrresponding
      headers. 
    """
    dataset = dataset_in.copy()

    # All frames must have the same CFAM filter
    cfam_list = []
    for frame in dataset:
        try:
            cfam_list += [frame.ext_hdr['CFAMNAME']]
        except:
            raise Exception('Frame w/o CFAM specification. All frames must have CFAM specified')
    if len(set(cfam_list)) != 1:
        raise Exception('All frames must have the same CFAM filter')

    # Get estimated PSF centers and CT
    psf_loc_est, ct_est = \
        corgidrp.corethroughput.estimate_psf_pix_and_ct(dataset,
            roi_radius=roi_radius,
            cfam_version=cfam_version)

    psf_hdu, dq_hdu = generate_psf_cube(dataset, psf_loc_est,
        cfam_name=cfam_list[0], cfam_version=cfam_version)
    # N sets of (x,y, CT measurements)
    # x, y: PSF centers wrt EXCAM's (0,0) pixel
    ct_excam = np.array([psf_loc_est[:,0], psf_loc_est[:,1], ct_est])
    ct_hdr = fits.Header()
    # Core throughput values on EXCAM wrt pixel (0,0) (not a "CT map", which is
    # wrt FPM's center 
    ct_hdr['COMMENT'] = ('PSF location with respect to EXCAM (0,0) pixel. '
        'Core throughput value for each PSF. (x,y,ct)=(data[0], data[1], data[2])')
    ct_hdr['UNITS'] = 'PSF location: EXCAM pixels. Core throughput: values between 0 and 1.'
    ct_hdu_list = [fits.ImageHDU(data=ct_excam, header=ct_hdr, name='CTEXCAM')]
    # Values of FPAM during CT observations (needed to derive the FPM's center
    # during CT observations given a coronagraphic dataset). The values do not
    # change during CT observations
    fpam_hv = [dataset_in[0].ext_hdr['FPAM_H'], dataset_in[0].ext_hdr['FPAM_V']]
    fpam_hdr = fits.Header()
    fpam_hdr['COMMENT'] = 'FPAM H and V values during the core throughput observations'
    fpam_hdr['UNITS'] = 'micrometer'
    ct_hdu_list += [fits.ImageHDU(data=fpam_hv, header=fpam_hdr, name='CTFPAM')]
    # Values of FSAM during CT observations (needed to derive the FPM's center
    # during CT observations given a coronagraphic dataset). The values do not
    # change during CT observations
    fsam_hv = [dataset_in[0].ext_hdr['FSAM_H'], dataset_in[0].ext_hdr['FSAM_V']]
    fsam_hdr = fits.Header()
    fsam_hdr['COMMENT'] = 'FSAM H and V values during the core throughput observations'
    fsam_hdr['UNITS'] = 'micrometer'
    ct_hdu_list += [fits.ImageHDU(data=fsam_hv, header=fsam_hdr, name='CTFSAM')]

    # Generate core throughput calibration file
    ct_cal = data.CoreThroughputCalibration(psf_hdu.data,
        pri_hdr=dataset[0].pri_hdr,
        ext_hdr=psf_hdu.header,
        input_hdulist=ct_hdu_list,
        dq=dq_hdu.data,
        dq_hdr=dq_hdu.header,
        input_dataset=dataset)

    return ct_cal

def get_1d_ct(ct_cal,frame,seps,
              method='nearest'):
    """Fetches core throughput values at specific separations from the mask center.
    Currently only the 'nearest' method is configured. 

    Args:
        ct_cal (corgidrp.data.CoreThroughputCalibration): the core throughput calibration 
            object.
        frame (corgidrp.data.Image): data frame containing mask location and detector 0,0 coordinate 
            in the header
        seps (np.array of float): separations (pixels from the mask center) at which to sample 
            the CT curve.
        method (str, optional): Method of calculating CT at a given separation. Defaults to 'nearest'.
            'nearest': grabs the core throughput measured at a location nearest to the desired 
            separation and assumes CT is radially symmetric.

    Returns:
        np.array: Array of shape (2,len(seps)), where the first row is the list of separations 
            sampled, and the second row is the ct value for each separation.
    """
    x, y, ct = ct_cal.ct_excam

    # Get location of mask center in CT coordinates
    xcen = frame.ext_hdr['STARLOCX'] + frame.ext_hdr.get("DETPIX0X",0.) + 0.5
    ycen = frame.ext_hdr['STARLOCY'] + frame.ext_hdr.get("DETPIX0Y",0.) + 0.5

    ct_seps = np.sqrt((x-xcen)**2 + (y-ycen)**2)

    if method == 'nearest':
        cts_out = []
        for sep in seps:
            argmin = np.argmin(np.abs(sep-ct_seps))
            ct_out = ct[argmin]
            cts_out.append(ct_out)
        
        ct_arr_out = np.array([seps,cts_out])
        return ct_arr_out
    else:
        raise NotImplementedError


def CreateCTMap(
    corDataset,
    fpamfsamcal,
    ct_cal,
    x_range=[-23,23],
    y_range=[-23,23],
    n_gridx=47,
    n_gridy=47,
    target_pix=None,
    logr=False,
    filepath=None,
    save=False):
    """
      Create a core throughput map: Given a core throughput calibration file and
      a coronagraphic dataset, derive 3-D list (x,y,ct) where (x,y) are some
      target locations on EXCAM relative to the FPM's center and with valid
      values of the throughput.

        The core throughmap may be saved, optionally, as a CSV file.

        The creation of the core throughput map relies on InterpolateCT(), a 
      method of the CoreThroughputCalibration class in data.py. Valid core
      throughput values are within the minimum and maxium radial distance from
      the FPM's center in the core throughput dataset used to generate the
      core throughput calibration file. Its options are inluded in the call of
      this method too.

      If an external list of locations is not provided, a default grid of points
      is condidered.

    Args:
      corDataset (corgidrp.data.Dataset): a dataset containing some
        coronagraphic observations.
      fpamfsamcal (corgidrp.data.FpamFsamCal): an instance of the
        FpamFsamCal class. That is, a FpamFsamCal calibration.
      ct_cal (corgidrp.data.CoreThroughputCalibration): an instance of the
        CoreThroughputCalibration class. That is, a core throughput calibration
        file.
      x_range (array): Two values [xmin, xmax] specifying the range of pixels to
        be considered. Units are EXCAM pixels measured with respect the center
        of the FPM. Notice that [-23,23] is approx. +/-10 l/D in band 1.
      y_range (array): Two values [xmin, xmax] specifying the range of pixels to
        be considered. Units are EXCAM pixels measured with respect the center
        of the FPM. Notice that [-23,23] is approx. +/-10 l/D in band 1.
      n_gridx (int) (optional): Number of x gridpoints.
      n_gridy (int) (optional): Number of y gridpoints.
      target_pix (array) (optional): a user-defined Mx2 array containing the pixel
        positions for M target pixels where the core throughput will be derived
        by interpolation. The target pixels are measured with respect the center
        of the focal plane mask in (fractional) EXCAM pixels. Default is None.
        In this case, a rectangular grid of pixel positions is used. Using
        matplotlib.pyplot, target_pix[0] is the horizontal axis (x), and
        target_pix[1] is the vertical axis (y).
      logr (bool) (optional): If True, radii are mapped into their logarithmic
        values before constructing the interpolant.
      filepath (string) (optional): String with the path and filename of the 
        file that will store the core throughput map as a CSV file.
      save (bool) (optionla): Whether the core throughput map will be stored or not.

    Returns:
        A core throughput map with (x,y,ct) where x and y are locations
        on EXCAM relative to the FPM's center with valid interpolated values of
        the core throughput.
    """
    # If no target pixels are provided, create a grid:
    if target_pix is None:
        x_tmp = np.linspace(x_range[0], x_range[1], n_gridx)
        y_tmp = np.linspace(y_range[0], y_range[1], n_gridy)
        target_pix = np.array(np.meshgrid(x_tmp, y_tmp)).reshape(2, n_gridx*n_gridy)
    # Get interpolated CT values at valid positions
    ct_interp = ct_cal.InterpolateCT(
            target_pix[0], target_pix[1], corDataset, fpamfsamcal, logr=logr)

    # Re-order output to match the required order: (x,y,ct)
    ct_map = ct_interp[[1,2,0]]
    
    # Optionally, store it
    if save:
        if filepath == None:
            # Choose a default location and filename
            filepath = os.path.join(corgidrp.default_cal_dir, 'ct_map.csv')
            np.savetxt(filepath, ct_map, delimiter=',', header='x, y, ct')
            print(f'CT map saved in {filepath:s}')
    
    return ct_map



#%%
import os
import glob
import pytest
import numpy as np
import corgidrp.data as data
import corgidrp.mocks as mocks
import corgidrp.detector as detector
import corgidrp.l2b_to_l3 as l2b_to_l3

#%%

def test_exptime_normalization():
    """
    Generate mock input data and pass into exposure_time_normalization function
    """
    ###### create simulated data
    # check that simulated data folder exists, and create if not
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    
    #create simulated data
    not_normalized_dataset = mocks.create_not_normalized_dataset(filedir=datadir)

    # extract the exposure time from each Image and check that it is >0
    exposure_times = np.zeros(len(not_normalized_dataset.frames))
    for i in range(len(not_normalized_dataset.frames)):
        assert not_normalized_dataset.frames[i].ext_hdr['EXPTIME'] > 0
        exposure_times[i] = not_normalized_dataset.frames[i].ext_hdr['EXPTIME']

    # divide the simulated data by the exposure time
    norm_dataset = l2b_to_l3.divide_by_exptime(not_normalized_dataset)

    # test if the function works
    for i in range(len(not_normalized_dataset.frames)):

        # test that the unit in the header is in electrons/s
        assert norm_dataset.frames[i].ext_hdr['BUNIT'] == "photoelectrons/s"

        #check that the quality flag has exists and it's 1 everywhere
        assert np.mean(not_normalized_dataset.frames[i].dq) == pytest.approx(0, abs=1e-6)

        # check that, for each frame, if you multiply the output by the exposure time you can recover the input
        assert not_normalized_dataset.frames[i].data == pytest.approx(exposure_times[i] * norm_dataset.frames[i].data, abs=1e-6)
        assert not_normalized_dataset.frames[i].err == pytest.approx(exposure_times[i] * norm_dataset.frames[i].err, abs=1e-6)

        # check that the output in .frames.data and .frames.err correspond to .all_data and .all_err
        assert norm_dataset.frames[i].data == pytest.approx(norm_dataset.all_data[i,:,:], abs = 1e-6)
        assert norm_dataset.frames[i].err == pytest.approx(norm_dataset.all_err[i,:,:], abs = 1e-6)

#%%    

if __name__ == "__main__":
    test_exptime_normalization()
# %%



import pytest

import numpy as np
from corgidrp.darks import mean_combine

ones = np.ones((10, 10))
zeros = np.zeros((10, 10), dtype=int)
vals = [1, 0, 4]
check_ims = []
check_masks = []
one_fr_mask = []
for val in vals:
    check_ims.append(ones*val)
    check_masks.append(zeros.copy())
    one_fr_mask.append(zeros.copy())
    pass

all_masks_i = (0, 0)
single_mask_i = (0, 1)
# Mask one pixel across all frames
for m in check_masks:
    m[all_masks_i[0], all_masks_i[1]] = 1
    pass

# Mask another pixel in only one frame
check_masks[0][single_mask_i[0], single_mask_i[1]] = 1
# Mask one pixel in one frame for one_fr_mask
one_fr_mask[0][single_mask_i[0], single_mask_i[1]] = 1


def test_mean_im():
    """Verify method calculates mean image and error term."""
    tol = 1e-13

    check_combined_im = np.mean(check_ims, axis=0)
    check_combined_im_err = np.sqrt(np.sum(np.array(check_ims)**2,
                                            axis=0)/len(check_ims))
    # For the pixel that is masked throughout
    check_combined_im[all_masks_i] = 0
    check_combined_im_err[all_masks_i] = 0
    unmasked_vals = np.delete(vals, single_mask_i[0])
    # For pixel that is only masked once
    check_combined_im[single_mask_i] = np.mean(unmasked_vals)
    check_combined_im_err[single_mask_i] = \
        np.sqrt(np.sum(unmasked_vals**2)/unmasked_vals.size)

    combined_im, _, _, _ = mean_combine(check_ims, check_masks)
    combined_im_err, _, _, _ = mean_combine(check_ims,
                                            check_masks, err=True)

    assert (np.max(np.abs(combined_im - check_combined_im)) < tol)
    assert (np.max(np.abs(combined_im_err - check_combined_im_err))
                                            < tol)


def test_mean_mask():
    """Verify method calculates correct mean mask."""
    _, combined_mask, _, _ = mean_combine(check_ims, check_masks)
    check_combined_mask = np.zeros_like(combined_mask)
    # Only this pixel should be combined
    check_combined_mask[all_masks_i] = 1

    assert ((combined_mask == check_combined_mask).all())


def test_darks_exception():
    """Half or more of the frames for a given pixel are masked."""
    # all frames for pixel (0,0) masked for inputs below
    _, _, _, enough_for_rn = mean_combine(check_ims, check_masks)
    assert (enough_for_rn == False)


def test_darks_map_im():
    """map_im as expected."""
    _, _, map_im, _  = mean_combine(check_ims,
                                            one_fr_mask)
    # 99 pixels with no mask on any of the 3 frames, one with one
    # frame masked
    # one pixel masked on one frame:
    assert(np.where(map_im == 2)[0].size == 1)
    # 99 pixels not masked on all 3 frames:
    assert(np.where(map_im == 3)[0].size == map_im.size - 1)
    expected_mean_num_good_fr = (3*99 + 2)/100
    assert (np.mean(map_im) == expected_mean_num_good_fr)


def test_invalid_image_list():
    """Invalid inputs caught"""

    bpmap_list = [np.zeros((3, 3), dtype=int), np.eye(3, dtype=int)]

    # for image_list
    perrlist = [
        'txt', None, 1j, 0, (5,), # not list
        (np.ones((3, 3)), np.ones((3, 3))), # not list
        [np.eye(3), np.eye(3), np.eye(3)], # length mismatch
        [np.eye(3)], # length mismatch
        [np.eye(4), np.eye(3)], # array size mismatch
        [np.eye(4), np.eye(4)], # array size mismatch
        [np.ones((1, 3, 3)), np.ones((1, 3, 3))], # not 2D
        ]

    for perr in perrlist:
        with pytest.raises(TypeError):
            mean_combine(perr, bpmap_list)
        pass

    # special empty case
    with pytest.raises(TypeError):
        mean_combine([], [])

    pass


def test_invalid_bpmap_list():
    """Invalid inputs caught"""

    image_list = [np.zeros((3, 3)), np.eye(3)]

    # for image_list
    perrlist = [
        'txt', None, 1j, 0, (5,), # not list
        (np.ones((3, 3), dtype=int), np.ones((3, 3), dtype=int)), #not list
        [np.eye(3, dtype=int), np.eye(3, dtype=int),
            np.eye(3, dtype=int)], # length mismatch
        [np.eye(3, dtype=int)], # length mismatch
        [np.eye(4, dtype=int), np.eye(3, dtype=int)], # array size mismatch
        [np.eye(4, dtype=int), np.eye(4, dtype=int)], # array size mismatch
        [np.ones((1, 3, 3), dtype=int),
            np.ones((1, 3, 3), dtype=int)], # not 2D
        [np.eye(3)*1.0, np.eye(3)*1.0], # not int
        ]

    for perr in perrlist:
        with pytest.raises(TypeError):
            mean_combine(image_list, perr)
        pass
    pass

def test_bpmap_list_element_not_0_or_1():
    """Catch case where bpmap is not 0 or 1"""

    image_list = [np.zeros((3, 3))]
    bpmap_list = [np.zeros((3, 3))]

    for val in [-1, 2]:
        bpmap_list[0][0, 0] = val
        with pytest.raises(TypeError):
            mean_combine(image_list, bpmap_list)
        pass
    pass

def test_accommodation_of_ndarray_inputs():
    """If inputs is a array_like (whether a single frame or a stack),
    the function will accommodate and convert each to a list of arrays."""

    # single frame
    image_list = np.ones((3,3))
    bpmap_list = np.zeros((3,3)).astype(int)
    # runs with no issues:
    mean_combine(image_list, bpmap_list)

    # stack
    image_list = np.stack([np.ones((3,3)), 2*np.ones((3,3))])
    bpmap_list = np.stack([np.zeros((3,3)).astype(int),
                            np.zeros((3,3)).astype(int)])
    # runs with no issues:
    mean_combine(image_list, bpmap_list)

if __name__ == '__main__':
    test_mean_im()
    test_mean_mask()
    test_darks_exception()
    test_darks_map_im()
    test_invalid_image_list()
    test_invalid_bpmap_list()
    test_bpmap_list_element_not_0_or_1()
    test_accommodation_of_ndarray_inputs()








import re
import os
import pytest
import pickle
import warnings
import numpy as np

from corgidrp.darks import (calibrate_darks_lsq,
            CalDarksLSQException)
from corgidrp.detector import slice_section, imaging_area_geom, imaging_slice
from corgidrp.mocks import create_synthesized_master_dark_calib
from corgidrp.mocks import detector_areas_test as dat
from corgidrp.data import DetectorNoiseMaps, DetectorParams, Dataset

# make test reproducible
np.random.seed(4567)
# use default parameters
detector_params = DetectorParams({})
# specified parameters simulated in simulated data from
# mocks.create_synthesized_master_dark_calib:
dark_current = 8.33e-4 #e-/pix/s
cic=0.02  # e-/pix/frame
rn=100 # e-/pix/frame
bias=2000 # e-
eperdn = 7 # e-/DN conversion; used in this example for all stacks

# EMgain_picks = (np.linspace(2, 5000, 7))
# exptime_picks = (np.linspace(2, 100, 7)) # 7x7 = 49 data points
# grid = np.meshgrid(EMgain_picks, exptime_picks)
# EMgain_arr = grid[0].ravel()
# exptime_arr = grid[1].ravel()
# kgain_arr = eperdn*np.ones_like(EMgain_arr) # all the same

#added in after emccd_detect makes the frames (see below)
FPN = 21 # e
# number of frames in each sub-stack of stack_arr:
N = 30#600 #30; can also replace with 30 to use those sub-stacks in the
#testdata_small folder; for 30, R_map will be smaller per pixel b/c of big
# variance, and F, C, and D will have more variance with respect to their
# expected values.  No need to test both here;  just N=30 case is sufficient.
# image area, including "shielded" rows and cols:
imrows, imcols, imr0c0 = imaging_area_geom('SCI', dat)

def setup_module():
    """
    Sets up testing module
    """
    global dataset
    dataset = create_synthesized_master_dark_calib(dat)

def teardown_module():
    """
    Runs at the end. Deletes big unused variables
    """
    global dataset
    del dataset

# filter out expected warnings
warnings.filterwarnings('ignore', category=UserWarning,
    module='corgidrp.darks')
# filter out expected warning when unreliable_pix_map has a pixel
# masked for all sub-stacks (which makes Rsq NaN from a division by 0)
warnings.filterwarnings('ignore', category=RuntimeWarning,
    module='corgidrp.darks')


def test_expected_results_sub():
    """Outputs are as expected, for smaller-sized frames."""

    noise_maps = calibrate_darks_lsq(dataset, detector_params, dat)
    F_image_map = imaging_slice('SCI', noise_maps.FPN_map, dat)
    # F
    assert (np.isclose(np.mean(F_image_map), FPN//eperdn*eperdn,
                                atol=FPN/5))
    # No FPN was inserted in non-image areas (so that bias subtraction
    #wouldn't simply remove it), so it should be 0 in prescan
    F_prescan_map = slice_section(noise_maps.FPN_map, 'SCI', 'prescan', dat)
    assert(np.isclose(np.nanmean(F_prescan_map), 0, atol=5))
    # C
    assert(np.isclose(np.nanmean(noise_maps.CIC_map), cic, atol=0.01))
    assert(len(noise_maps.CIC_map[noise_maps.CIC_map < 0]) == 0)
    # D
    D_image_map = imaging_slice('SCI', noise_maps.DC_map, dat)
    assert(np.isclose(np.mean(D_image_map),
                    dark_current, atol=2e-4))
    assert(len(noise_maps.DC_map[noise_maps.DC_map < 0]) == 0)
    # D_map: 0 everywhere except image area
    im_rows, im_cols, r0c0 = imaging_area_geom('SCI', dat)
    # D_nonimg = D_map[~D_map[r0c0[0]:r0c0[0]+im_rows,
    #             r0c0[1]:r0c0[1]+im_cols]]
    D_map = noise_maps.DC_map.copy()
    D_map[r0c0[0]:r0c0[0]+im_rows,
                r0c0[1]:r0c0[1]+im_cols] = 0
    # now whole map should be 0
    assert(np.nanmin(D_map) == 0)
    if N == 30:
        # bias_offset (tolerance of 5 for N=30 since I used a small number
        # of frames for that dataset, especially for the high-gain ones)
        assert(np.isclose(noise_maps.bias_offset, 0, atol=5)) #in DN
    if N == 600:
        assert(np.isclose(noise_maps.bias_offset, 0, atol=1)) #in DN
    # dark current only in image area
    D_std_map_im = noise_maps.DC_err[r0c0[0]:r0c0[0]+im_rows,
                                r0c0[1]:r0c0[1]+im_cols]
    # assert that the std dev coming from the fit itself is < noise itself
    assert(np.nanmean(D_std_map_im) < np.nanmean(D_image_map))
    # noise_maps.CIC_err accounts for the err from the input frames and the
    # statistical error (std dev across frames), so skip the assertion for CIC here
    assert(np.nanmean(noise_maps.FPN_err) < np.nanmean(noise_maps.FPN_map))

    # check the noisemap can be pickled (for CTC operations)
    pickled = pickle.dumps(noise_maps)
    pickled_noisemap = pickle.loads(pickled)
    assert np.all((noise_maps.data == pickled_noisemap.data) | np.isnan(noise_maps.data))

    # save noise map
    calibdir = os.path.join(os.path.dirname(__file__), "testcalib")
    nm_filename = noise_maps.filename
    if not os.path.exists(calibdir):
        os.mkdir(calibdir)
    noise_maps.save(filedir=calibdir, filename=nm_filename)
    nm_filepath = os.path.join(calibdir, nm_filename)
    nm_f = DetectorNoiseMaps(nm_filepath)

    # check the noisemap can be pickled (for CTC operations)
    pickled = pickle.dumps(nm_f)
    pickled_noisemap = pickle.loads(pickled)
    assert np.all((noise_maps.data == pickled_noisemap.data) | np.isnan(noise_maps.data))

    # tests the copy method, from filepath way of creating class
    # instance, too
    nm_open = nm_f.copy()
    assert(np.array_equal(nm_open.data, noise_maps.data, equal_nan=True))

    # check headers
    assert(noise_maps.data.ndim == 3)
    test_filename = dataset.frames[-1].filename.split('.fits')[0] + '_DNM_CAL.fits'
    test_filename = re.sub('_L[0-9].', '', test_filename)
    assert(noise_maps.filename == test_filename)
    assert(noise_maps.ext_hdr["BUNIT"] == "Detected Electrons")
    assert(noise_maps.err_hdr["BUNIT"] == "Detected Electrons")
    assert("DetectorNoiseMaps" in str(noise_maps.ext_hdr["HISTORY"]))
    assert(noise_maps.ext_hdr['B_O_UNIT'] == 'DN')

    assert(nm_open.data.ndim == 3)
    assert(nm_open.filename == test_filename)
    assert(nm_open.ext_hdr["BUNIT"] == "Detected Electrons")
    assert(nm_open.err_hdr["BUNIT"] == "Detected Electrons")
    assert("DetectorNoiseMaps" in str(nm_open.ext_hdr["HISTORY"]))
    assert(nm_open.ext_hdr['B_O_UNIT'] == 'DN')

def test_sub_stack_len():
    """datasets should have at least 4 sub-stacks."""
    data_set = dataset.copy()
    ds, _ = data_set.split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR'])
    # make a dataset with only 2 distinct combos
    dataset_few = Dataset([ds[0].frames[k] for k in range(len(ds[0]))] + [ds[1].frames[k] for k in range(len(ds[1]))])
    with pytest.raises(CalDarksLSQException):
        calibrate_darks_lsq(dataset_few, detector_params, dat)

def test_g_arr_unique():
    '''EM gains must have at least 2 unique elements.'''
    data_set = dataset.copy()
    ds, _ = data_set.split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR'])
    for j in range(len(ds)):
        for d in ds[j].frames:
            d.ext_hdr['EMGAIN_C'] = 4
    with pytest.raises(CalDarksLSQException):
        calibrate_darks_lsq(data_set, detector_params, dat)

def test_g_gtr_1():
    '''EM gains must all be bigger than 1.'''
    data_set = dataset.copy()
    ds, _ = data_set.split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR'])
    ds[0].frames[0].ext_hdr['EMGAIN_C'] = 1
    with pytest.raises(CalDarksLSQException):
        calibrate_darks_lsq(data_set, detector_params, dat)

def test_t_arr_unique():
    '''Exposure times must have at least 2 unique elements.'''
    data_set = dataset.copy()
    ds, _ = data_set.split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR'])
    for j in range(len(ds)):
        for d in ds[j].frames:
            d.ext_hdr['EXPTIME'] = 4
    with pytest.raises(CalDarksLSQException):
        calibrate_darks_lsq(data_set, detector_params, dat)

def test_t_gtr_0():
    '''Exposure times must all be bigger than 0.'''
    data_set = dataset.copy()
    ds, _ = data_set.split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR'])
    ds[0].frames[0].ext_hdr['EXPTIME'] = 0
    with pytest.raises(CalDarksLSQException):
        calibrate_darks_lsq(data_set, detector_params, dat)


def test_k_gtr_0():
    '''K gains must all be bigger than 0.'''
    data_set = dataset.copy()
    ds, _ = data_set.split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR'])
    ds[0].frames[0].ext_hdr['KGAINPAR'] = 0
    with pytest.raises(CalDarksLSQException):
        calibrate_darks_lsq(data_set, detector_params, dat)

def test_mean_num():
    '''If too many masked in a stack for a given pixel, warning raised. Checks
    that dq values are as expected, too.
    '''
    data_set = dataset.copy()

    ds, _ = data_set.split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR'])
    # tag 48 of the 49 sub-stacks as bad pixel all the
    # way through for one pixel (7,8)
    # And mask (10,12) to get flag value of 256
    for i in range(48):
        ds[i].all_dq[:,7,8] = 4
        ds[i].all_dq[:int(1+len(ds[i])/2),10,12] = 2

    with pytest.warns(UserWarning):
        nm_out = calibrate_darks_lsq(data_set, detector_params, dat)
    # last of out is the DetectorNoiseMaps instance
    # And dq is really a 3-frame stack, and all 3 are the same.  So pick one of them.
    assert nm_out.dq[0,7,8] == 1
    assert nm_out.dq[0,10,12] == 256


if __name__ == '__main__':
    setup_module()

    test_mean_num()
    test_expected_results_sub()
    test_sub_stack_len()
    test_g_arr_unique()
    test_g_gtr_1()
    test_t_arr_unique()
    test_t_gtr_0()
    test_k_gtr_0()

    teardown_module()






import os
import glob
import pickle
import pytest
import numpy as np
import corgidrp
import corgidrp.data as data
import corgidrp.mocks as mocks
import corgidrp.detector as detector
import corgidrp.l2a_to_l2b as l2a_to_l2b
from corgidrp.darks import build_trad_dark

np.random.seed(456)

old_err_tracking = corgidrp.track_individual_errors
# use default parameters
detector_params = data.DetectorParams({})

# make a mock DetectorNoiseMaps instance (to get the bias offset input)
im_rows, im_cols, _ = detector.unpack_geom('SCI', 'image')
rows = detector.detector_areas['SCI']['frame_rows']
cols = detector.detector_areas['SCI']['frame_cols']

Fd = np.ones((rows, cols))
Dd = 3/3600*np.ones((rows, cols))
Cd = 0.02*np.ones((rows, cols))

Ferr = np.zeros((rows, cols))
Derr = np.zeros((rows, cols))
Cerr = np.zeros((rows, cols))
Fdq = Ferr.copy().astype(int)
Ddq = Derr.copy().astype(int)
Cdq = Cerr.copy().astype(int)
noise_maps = mocks.create_noise_maps(Fd, Ferr, Fdq, Cd,
                                            Cerr, Cdq, Dd, Derr, Ddq)

def test_dark_sub():
    """
    Generate mock input data and pass into dark subtraction function
    """
    corgidrp.track_individual_errors = True # this test uses individual error components

    ###### create simulated data
    # check that simulated data folder exists, and create if not
    datadir = os.path.join(os.path.dirname(__file__), "simdata")
    if not os.path.exists(datadir):
        os.mkdir(datadir)

    mocks.create_dark_calib_files(filedir=datadir)

    ####### test data architecture
    dark_filenames = glob.glob(os.path.join(datadir, "simcal_dark*.fits"))

    dark_dataset = data.Dataset(dark_filenames)

    assert len(dark_dataset) == 10

    # check that data is consistently modified
    dark_dataset.all_data[0,0,0] = 0
    assert dark_dataset[0].data[0,0] == 0

    dark_dataset[0].data[0,0] = 1
    assert dark_dataset.all_data[0,0,0] == 1

    ###### create dark
    dark_frame = build_trad_dark(dark_dataset, detector_params, detector_regions=None, full_frame=True)

    # check the level of dark current is approximately correct; leave off last row, telemetry row
    assert np.mean(dark_frame.data[:-1]) == pytest.approx(150, abs=1e-2)

    # leave out telemetry row of full frame
    dark_dataset_notelem = dark_dataset.all_data[:, 0:-1, :]
    assert np.array_equal(np.std(dark_dataset_notelem, axis = 0)/np.sqrt(len(dark_dataset_notelem)), dark_frame.err[0][0:-1, :])

    # save dark
    calibdir = os.path.join(os.path.dirname(__file__), "testcalib")
    dark_filename = "sim_dark_calib.fits"
    if not os.path.exists(calibdir):
        os.mkdir(calibdir)
    dark_frame.save(filedir=calibdir, filename=dark_filename)

    ###### perform dark subtraction
    # load in the dark
    dark_filepath = os.path.join(calibdir, dark_filename)
    new_dark = data.Dark(dark_filepath)
    assert new_dark.ext_hdr['PC_STAT'] == 'analog master dark'

    # check the dark can be pickled (for CTC operations)
    pickled = pickle.dumps(new_dark)
    pickled_dark = pickle.loads(pickled)
    assert np.all(new_dark.data == pickled_dark.data)

    # subtract darks from itself; also saves to testcalib folder
    darkest_dataset = l2a_to_l2b.dark_subtraction(dark_dataset, new_dark, outputdir=calibdir)
    assert(dark_filename in str(darkest_dataset[0].ext_hdr["HISTORY"]))

    # PC dark cannot be used for this step function
    new_dark.ext_hdr['PC_STAT'] = 'photon-counted master dark'
    with pytest.raises(Exception):
        l2a_to_l2b.dark_subtraction(dark_dataset, new_dark, outputdir=calibdir)

    # check the level of the dataset is now approximately 0, leaving off telemetry row
    assert np.mean(darkest_dataset.all_data[:,:-1,:]) == pytest.approx(0, abs=1e-2)

    # check that image area option works
    dark_frame = build_trad_dark(dark_dataset, detector_params, detector_regions=None, full_frame=False)
    # compare, but first get image area from input:
    dark_dataset_im = dark_dataset.copy()
    dark_im_frames = []
    for fr in dark_dataset_im.all_data:
        fr = detector.slice_section(fr, 'SCI', 'image')
        dark_im_frames.append(fr)
    dark_im_frames = np.stack(dark_im_frames)
    # check that the error is determined correctly (no telemetry row in image area)
    assert np.array_equal(np.std(dark_im_frames, axis = 0)/np.sqrt(len(dark_im_frames)), dark_frame.err[0])

    # check the propagated errors
    assert darkest_dataset[0].err_hdr["Layer_2"] == "dark_error"
    assert(np.mean(darkest_dataset.all_err) == pytest.approx(np.mean(dark_frame.err), abs = 1e-2))
    #print("mean of all data:", np.mean(darkest_dataset.all_data))
    #print("mean of all errors:", np.mean(darkest_dataset.all_err))
    assert darkest_dataset[0].ext_hdr["BUNIT"] == "Photoelectrons"
    assert darkest_dataset[0].err_hdr["BUNIT"] == "Photoelectrons"
    #print(darkest_dataset[0].ext_hdr)

    # If too many masked in a stack for a given pixel, warning raised. Checks
    # that dq values are as expected, too.
    ds = dark_dataset.copy()
    # tag as bad pixel all the
    # way through for one pixel (7,8)
    # And mask (10,12) to get flag value of 256
    ds.all_dq[:,7,8] = 4
    ds.all_dq[:int(1+len(ds)/2),10,12] = 2

    with pytest.warns(UserWarning):
        master_dark = build_trad_dark(ds, detector_params, full_frame=True)
    assert master_dark.dq[7,8] == 1
    assert master_dark.dq[10,12] == 256

    # now input a DetectorNoiseMaps instance and subtract dark from itself; here outputdir will do nothing since this is not a Dark instance
    EMgain = 10
    exptime = 4
    frame = (noise_maps.FPN_map + noise_maps.CIC_map*EMgain + noise_maps.DC_map*exptime*EMgain)/EMgain
    prihdr, exthdr = mocks.create_default_calibration_product_headers()
    image_frame = data.Image(frame, prihdr, exthdr)
    image_frame.ext_hdr['EMGAIN_C'] = EMgain
    image_frame.ext_hdr['EXPTIME'] = exptime
    image_frame.ext_hdr['KGAINPAR'] = 7
    dataset_from_noisemap = data.Dataset([image_frame])
    nm_dataset0 = l2a_to_l2b.dark_subtraction(dataset_from_noisemap, noise_maps, outputdir=calibdir)
    # check the level of the dataset is now approximately 0, leaving off telemetry row
    assert np.mean(nm_dataset0.all_data[:,:-1,:]) == pytest.approx(0, abs=1e-2)

    # check the dark can be pickled (for CTC operations)
    pickled = pickle.dumps(master_dark)
    pickled_dark = pickle.loads(pickled)
    assert np.all(master_dark.data == pickled_dark.data)

    corgidrp.track_individual_errors = old_err_tracking

if __name__ == "__main__":
    test_dark_sub()



import os
import re
import math
import numpy as np
from astropy.io import fits
import corgidrp.fluxcal as fluxcal
import corgidrp.astrom as astrom
from corgidrp.data import Dataset, FluxcalFactor, NDFilterSweetSpotDataset
from corgidrp.astrom import centroid_with_roi
from scipy.interpolate import griddata
import warnings

# =============================================================================
# Helper Functions
# =============================================================================

def load_transformation_matrix_from_fits(file_path):
    """
    Load a transformation matrix from a FITS file.
    
    Parameters:
        file_path (str): Path to the FITS file containing the transformation matrix.
    
    Returns:
        np.ndarray: The transformation matrix extracted from the FITS file.
    """
    with fits.open(file_path) as hdul:
        data = hdul[0].data
        if data is None:
            data = hdul[1].data
        return np.array(data)

def group_by_keyword(dataset, prihdr_keyword=None, exthdr_keyword=None):
    """
    Split the dataset by either a primary header (prihdr) or extension header (exthdr) keyword
    and return a dictionary {target: subset}.

    Parameters:
        dataset (Dataset): The dataset to be split.
        prihdr_keyword (str, optional): FITS primary header keyword to split the dataset on.
        exthdr_keyword (str, optional): FITS extension header keyword to split the dataset on.

    Returns:
        dict: A dictionary where keys are unique target values and values are the 
            corresponding dataset subsets.
    
    Raises:
        ValueError: If neither keyword is provided.
    """
    if not prihdr_keyword and not exthdr_keyword:
        raise ValueError("At least one of 'prihdr_keyword' or 'exthdr_keyword' must be provided.")

    # Determine the splitting method
    if prihdr_keyword:
        split_datasets, unique_vals = dataset.split_dataset(prihdr_keywords=[prihdr_keyword])
    else:
        split_datasets, unique_vals = dataset.split_dataset(exthdr_keywords=[exthdr_keyword])

    # Construct dictionary {target: subset}
    groups = {}
    for key, sub_ds in zip(unique_vals, split_datasets):
        target = key[0] if isinstance(key, tuple) else key
        groups[target] = sub_ds

    return groups


def interpolate_od(sweet_spot_data, x_query, y_query, method='linear'):
    """
    Interpolate OD from sweet_spot_data at (x_query, y_query).

    Parameters:
        sweet_spot_data (numpy.ndarray): An Nx3 array where each row contains [OD, x, y].
        x_query (float or numpy.ndarray): The x-coordinate(s) at which to interpolate.
        y_query (float or numpy.ndarray): The y-coordinate(s) at which to interpolate.
        method (str): The interpolation method to use ('linear', 'nearest', or 'cubic'). 
            Defaults to 'linear'.

    Returns:
        float or numpy.ndarray: The interpolated OD value(s). Returns a float if a single 
            query point is provided, otherwise returns an array of interpolated values.
    """

    # Points and OD values
    points = sweet_spot_data[:, 1:3]  # shape (N,2) -> (x, y)
    values = sweet_spot_data[:, 0]    # shape (N,) -> OD

    # Prepare interpolation coordinates
    if np.isscalar(x_query) and np.isscalar(y_query):
        xi = (x_query, y_query)
    else:
        xi = np.column_stack((x_query, y_query))

    od_interp = griddata(points, values, xi, method=method)
    return float(od_interp) if np.isscalar(x_query) else od_interp


# =============================================================================
# Flux Calibration Helpers
# =============================================================================

def compute_expected_band_irradiance(star_name, filter_name):
    """
    Compute the expected band-integrated irradiance (erg/(s*cm^2)) for a given star.

    Parameters:
        star_name (str): The name of the star for which to compute the irradiance.
        filter_name (str): The name of the filter used to determine the transmission curve.

    Returns:
        float: The computed band-integrated irradiance in erg/(s*cm^2).
    
    Raises:
        ValueError: If no matching filter curve file is found.
    """
    calspec_filepath = fluxcal.get_calspec_file(star_name)
    datadir = os.path.join(os.path.dirname(fluxcal.__file__), "data", "filter_curves")
    filter_files = [f for f in os.listdir(datadir) if filter_name in f and f.endswith('.csv')]
    if not filter_files:
        raise ValueError(f"No filter curve available with name {filter_name}")
    
    filter_filename = os.path.join(datadir, filter_files[0])
    wave, transmission = fluxcal.read_filter_curve(filter_filename)
    calspec_flux = fluxcal.read_cal_spec(calspec_filepath, wave)
    return fluxcal.calculate_band_irradiance(transmission, calspec_flux, wave)


def compute_avg_calibration_factor(dim_stars_dataset, phot_method, flux_or_irr="irr", phot_kwargs=None):
    """
    Compute the average flux calibration factor using dim stars (no ND filter).

    Parameters:
        dim_stars_dataset (iterable): Dataset containing dim star entries.
        phot_method (str): Photometry method to use ("Aperture" or "Gaussian").
        flux_or_irr (str): Whether flux ('flux') or in-band irradiance ('irr') should be used.
        phot_kwargs (dict, optional): Dictionary of keyword arguments to pass to calibrate_fluxcal_aper.

    Returns:
        float: The average calibration factor.
    """
    if phot_kwargs is None:
        phot_kwargs = {}

    if phot_method == "Aperture":
        cal_values = [
            fluxcal.calibrate_fluxcal_aper(entry, flux_or_irr, phot_kwargs).fluxcal_fac
            for entry in dim_stars_dataset
        ]
    elif phot_method == "Gaussian":
        cal_values = [
            fluxcal.calibrate_fluxcal_gauss2d(entry, flux_or_irr, phot_kwargs).fluxcal_fac
            for entry in dim_stars_dataset
        ]
    else:
        raise ValueError("Photometry method must be either Aperture or Gaussian.")

    return np.mean(cal_values)


# =============================================================================
# OD & Photometry Computations
# =============================================================================

def _compute_od_for_file(entry, target, phot_method, phot_kwargs, ref_fpam_name,
                         ref_fpam_h, ref_fpam_v, ref_cfam_name, expected_flux):
    """
    Helper subfunction to:
      1. Validate FPAM/CFAM metadata vs. the reference.
      2. Compute centroid (x, y).
      3. Perform photometry (Aperture or Gaussian).
      4. Compute OD from measured flux and the expected flux.
    
    Parameters:
        entry (corgidrp.Data.Image): The dataset entry containing image data and metadata.
        target (str): The target identifier for the dataset entry.
        phot_method (str): The photometry method to use ('Aperture' or 'Gaussian').
        phot_kwargs (dict): Additional keyword arguments for the photometry method.
        ref_fpam_name (str): The reference FPAM name for validation.
        ref_fpam_h (float): The reference FPAM horizontal position.
        ref_fpam_v (float): The reference FPAM vertical position.
        ref_cfam_name (str): The reference CFAM name for validation.
        expected_flux (float): The expected flux value for computing OD in erg/(s*cm^2)

    Returns:
        tuple:
            float: The computed optical depth (OD).
            float: The x-coordinate of the centroid.
            float: The y-coordinate of the centroid.

    Raises:
        ValueError: If FPAM/CFAM metadata do not match the reference values.
        ValueError: If an invalid photometry method is specified.
    """
    hdr = entry.ext_hdr

    # Metadata checks
    if (hdr.get('FPAMNAME') != ref_fpam_name or 
        abs(hdr.get('FPAM_H') - ref_fpam_h) > 1.2 or    # within non-repeatability tolerance of 1.2 um
        abs(hdr.get('FPAM_V') - ref_fpam_v) > 1.2 or 
        hdr.get('CFAMNAME') != ref_cfam_name):
        raise ValueError(
            f"Inconsistent FPAM/CFAM header values in target {target} for file {entry}!"
        )

    # Centroid
    x_center, y_center = centroid_with_roi(entry.data)
    if np.isnan(x_center) or np.isnan(y_center):
        print(f"Warning: Centroid could not be computed for {entry}")
        return None, None, None

    # Photometry
    if phot_method == "Aperture":
        phot_result = fluxcal.aper_phot(entry, **phot_kwargs)
    elif phot_method == "Gaussian":
        phot_result = fluxcal.phot_by_gauss2d_fit(entry, **phot_kwargs)
    else:
        raise ValueError("phot_method must be Aperture or Gaussian.")

    # Compute OD
    transmission = phot_result[0] / expected_flux
    od = -math.log10(transmission)
    return od, x_center, y_center


def process_bright_target(target, files, cal_factor, od_raster_threshold,
                          phot_method="Aperture", phot_kwargs=None):
    """
    Process bright star files for one target to compute optical density (OD)
    and (x, y) centroids for each dithered observation.
    Checks that FPAM keywords are consistent across all files.
    
    Additional photometry options are passed via phot_kwargs.
    This allows users to override default settings for functions like aper_phot.
    
    Parameters:
        target (str): The target star name.
        files (corgidrp.data.Dataset): Dataset of bright star images
        cal_factor (float or corgidrp.data.FluxcalFactor): Calibration factor.
        od_raster_threshold (float): Threshold for flagging OD variations.
        phot_method (str): Photometry method to use ("Aperture" or "Gaussian").
        phot_kwargs (dict, optional): Dictionary of keyword arguments to forward to the photometry function.
    
    Returns:
        dict: A dictionary containing computed OD values, centroids, and other metadata.
    """
    if phot_kwargs is None:
        phot_kwargs = {}

    first_hdr = files[0].ext_hdr
    ref_cfam_name = first_hdr['CFAMNAME']
    common_fpam_name = first_hdr.get('FPAMNAME')
    common_fpam_h    = first_hdr.get('FPAM_H')
    common_fpam_v    = first_hdr.get('FPAM_V')
    exptime          = first_hdr.get('EXPTIME')

    if type(cal_factor) == FluxcalFactor:
        cal_factor_value = cal_factor.fluxcal_fac
    else:
        cal_factor_value = cal_factor

    # Compute expected flux
    expected_irradiance_no_nd = compute_expected_band_irradiance(target, ref_cfam_name)
    expected_flux = expected_irradiance_no_nd / cal_factor_value

    od_values, x_values, y_values = [], [], []

    for entry in files:
        od, x_center, y_center = _compute_od_for_file(entry, target, phot_method, 
                                                      phot_kwargs, common_fpam_name, 
                                                      common_fpam_h, common_fpam_v, 
                                                      ref_cfam_name, expected_flux)
        
        # Skip if centroid was not valid
        if od is None:
            continue

        od_values.append(od)
        x_values.append(x_center)
        y_values.append(y_center)

    od_array = np.array(od_values)

    # Check for wide variation in OD values
    # Check for wide variation in OD values
    if od_array.size > 0:
        od_std = np.std(od_array)
        if od_std >= od_raster_threshold:
            warnings.warn(
                f"OD variation is high for target '{target}': "
                f"Standard deviation ({od_std:.3f}) exceeds threshold ({od_raster_threshold:.3f})."
            )
    else:
        od_std = np.nan

    average_od = np.mean(od_array) if od_array.size > 0 else np.nan

    return {
        'od_values': od_array,
        'average_od': average_od,
        'FPAMNAME': common_fpam_name,
        'FPAM_H': common_fpam_h,
        'FPAM_V': common_fpam_v,
        'CFAMNAME': ref_cfam_name,
        'flag': (od_std >= od_raster_threshold if not np.isnan(od_std) else False),
        'x_values': x_values,
        'y_values': y_values
    }


def create_nd_sweet_spot_dataset(aggregated_sweet_spot_data, common_metadata, od_var_flag, 
                                 input_dataset):
    """
    Create an NDFilterSweetSpotDataset FITS file with the Nx3 sweet-spot array.
    
    Parameters:
        aggregated_sweet_spot_data (numpy.ndarray): The aggregated Nx3 array containing 
            sweet-spot data in the format [OD, x, y].
        common_metadata (dict): A dictionary containing metadata such as FPAM/CFAM names 
            and offsets.
        od_var_flag (Bool): A flag that is passed in if the OD variance is too high among 
            rasters.
        input_dataset (corgidrp.data.Dataset): input dataset used to create the ND Filter 
            calibration

    Returns:
        tuple:
            NDFilterSweetSpotDataset: The generated ND filter sweet spot dataset.
    """
    final_sweet_spot_data = aggregated_sweet_spot_data.copy()

    # Build the NDFilterSweetSpotDataset, use info from last/ most recent
    pri_hdr = input_dataset[-1].pri_hdr
    ext_hdr = input_dataset[-1].ext_hdr

    # keeping the common metadata because if you do provide dim stars as part of the dataset
    # and grab the last header you might get the FPAM info of a frame with no ND filter in.
    ext_hdr['BUNIT']    = 'None (dimensionless)'
    ext_hdr['DATALVL']  = 'CAL'
    ext_hdr['FPAMNAME'] = common_metadata.get('FPAMNAME')
    ext_hdr['FPAM_H']   = common_metadata.get('FPAM_H')
    ext_hdr['FPAM_V']   = common_metadata.get('FPAM_V')
    ext_hdr['ODFLAG'] = od_var_flag
    ext_hdr['HISTORY']  = "Combined sweet-spot dataset from bright star dithers"

    ndsweetspot_dataset = NDFilterSweetSpotDataset(
        data_or_filepath=final_sweet_spot_data,
        pri_hdr=pri_hdr,
        ext_hdr=ext_hdr,
        input_dataset=input_dataset
    )

    return ndsweetspot_dataset


def calculate_od_at_new_location(clean_frame_entry, transformation_matrix_file, 
                                 ndsweetspot_dataset):
    """
    Use the NDFilterSweetSpot Dataset to calculate the OD at a new location for an input 
    image, using a provided EXCAM to FPAM transformation_matrix.
    
    Parameters:
        clean_frame_entry (corgidrp.Data.Image): A clean frame image.
        transformation_matrix_file (string): File path to the 2x2 matrix for transforming 
            FPAM offsets to EXCAM offsets.
            # TO DO: is this going to be a data object?
        ndsweetspot_dataset (corgidrp.Data.NDFilterSweetSpotDataset): ND Filter 
            Sweet Spot dataset

    Returns:
        interpolated_od (float): OD that is interpolated at the new star location
    """
    final_sweet_spot_data = ndsweetspot_dataset.data
    transformation_matrix = load_transformation_matrix_from_fits(transformation_matrix_file)

    if (clean_frame_entry is not None) and (transformation_matrix is not None):
        x_clean, y_clean = centroid_with_roi(clean_frame_entry.data)
        cframe_hdr = clean_frame_entry.ext_hdr
        sweetspot_hdr = ndsweetspot_dataset.ext_hdr

        # Compute FPAM offset
        clean_fpam_h = cframe_hdr.get('FPAM_H', 0.0)
        clean_fpam_v = cframe_hdr.get('FPAM_V', 0.0)
        sp_fpam_h    = sweetspot_hdr.get('FPAM_H', 0.0)
        sp_fpam_v    = sweetspot_hdr.get('FPAM_V', 0.0)
        fpam_offset  = np.array([clean_fpam_h - sp_fpam_h, clean_fpam_v - sp_fpam_v])

        # Transform to EXCAM offset
        excam_offset = transformation_matrix @ fpam_offset
        x_adj = x_clean + excam_offset[0]
        y_adj = y_clean + excam_offset[1]

        # Interpolate OD at that new location
        interpolated_od = interpolate_od(final_sweet_spot_data, x_adj, y_adj)

    # TO DO: add in interpolated od into the header of the file and re-save? determine how the OD 
    # will be propagated

    return interpolated_od

# =============================================================================
# Main Workflow Function
# =============================================================================

def create_nd_filter_cal(stars_dataset,
                         od_raster_threshold = 0.1,
                         phot_method="Aperture",
                         flux_or_irr="irr",
                         phot_kwargs=None,
                         fluxcal_factor=None):
    """
    Main ND Filter calibration workflow:
      1. Split dataset into dim and bright stars based on FPAMNAME keyword (or use cal factor input for dim)
      2. Compute avg calibration factor from dim stars.
      2. Group bright star frames by target + measure OD, centroids.
      3. Combine all sweet-spot data into a single Nx3 array.
    
    Parameters:
        stars_dataset (Dataset): Dataset containing star images. The splitting into bright and dim stars
            is performed based on the 'FPAMNAME' value in the FITS header. For example, entries with 'FPAMNAME'
            containing "dim" (case-insensitive) are considered dim stars.
        od_raster_threshold (float): Threshold for flagging OD variations.
            # TO DO: figure out what a reasonable value for this should be 
        phot_method (str): Photometry method ("Aperture" or "Gaussian").
        flux_or_irr (str): Either 'flux' or 'irr' for the calibration approach.
        phot_kwargs (dict, optional): Extra arguments for the actual photometry function 
            (e.g., aper_phot).
        fluxcal_factor (corgidrp.Data.FluxcalFactor, optional): A pre-computed flux factor calibration product to use
            if dim stars are not included as part of the input dataset

    Returns:
        sweet_spot_dataset (corgidrp.Data.NDFilterSweetSpotDataset): ND Filter calibration product for the dataset given
    """
    if phot_kwargs is None:
        phot_kwargs = {}

    # 1. Split the stars dataset into dim and bright stars based on FPAMNAME or FSAMNAME
    dim_stars_dataset = []
    bright_stars_dataset = []
    try:
        grouped_nd_files = group_by_keyword(stars_dataset, prihdr_keyword=None, exthdr_keyword='FPAMNAME')
    except:
        grouped_nd_files = group_by_keyword(stars_dataset, prihdr_keyword=None, exthdr_keyword='FSAMNAME')
        
    dim_stars_dataset = []
    bright_stars_dataset = []

    for keyword, records in grouped_nd_files.items():
        if keyword.startswith('ND'):
            # don't overwrite
            bright_stars_dataset.extend(records)
        else:
            dim_stars_dataset.extend(records)

    bright_stars_dataset = Dataset(bright_stars_dataset)
    dim_stars_dataset = Dataset(dim_stars_dataset)


    # 2. If a fluxcal factor was provided, use that for the dim stars
    if fluxcal_factor is not None:
        cal_factor = fluxcal_factor
    else:
        # Otherwise, compute the average calibration factor from dim
        # star frames
        cal_factor = compute_avg_calibration_factor(dim_stars_dataset,
                                                    phot_method,
                                                    flux_or_irr,
                                                    phot_kwargs)

    # 3. Process bright star frames
    grouped_files = group_by_keyword(bright_stars_dataset, prihdr_keyword='TARGET', exthdr_keyword=None)
    flux_results = {}
    aggregated_data_list = []
    common_metadata = {}

    for target, files in grouped_files.items():
        if not files:
            continue
        print(f"Processing bright target files: {target}")
        star_data = process_bright_target(target, files, cal_factor,
                                          od_raster_threshold, phot_method,
                                          phot_kwargs)
        flux_results[target] = star_data

        od_var_flag = star_data['flag']

        # Convert to Nx3 array [OD, x, y]
        target_sweet_spot = np.column_stack((
            star_data['od_values'],
            star_data['x_values'],
            star_data['y_values']
        ))
        aggregated_data_list.append(target_sweet_spot)

        # Initialize or validate the common metadata
        if not common_metadata:
            common_metadata = {
                'FPAMNAME': star_data['FPAMNAME'],
                'FPAM_H': star_data['FPAM_H'],
                'FPAM_V': star_data['FPAM_V'],
                'CFAMNAME': star_data['CFAMNAME']
            }
        else:
            # Basic consistency checks
            if (common_metadata['FPAMNAME'] != star_data['FPAMNAME']
                or abs(common_metadata['FPAM_H'] - star_data['FPAM_H']) >= 1.2  # PAM non-repeatability tolerance is +/- 1.2 um
                or abs(common_metadata['FPAM_V'] - star_data['FPAM_V']) >= 1.2
                or common_metadata['CFAMNAME'] != star_data['CFAMNAME']):
                raise ValueError("Inconsistent FPAM or filter metadata among bright star observations.")

    # 4. Combine all sweet-spot arrays into one dataset
    combined_sweet_spot_data = (
        np.vstack(aggregated_data_list) if aggregated_data_list else np.empty((0, 3))
    )
    od_list = [res['average_od'] for res in flux_results.values()
               if res.get('average_od') is not None]
    overall_avg_od = np.mean(od_list) if od_list else None
    print(f"Average OD across bright targets: {overall_avg_od}")

    # 5. Create the final NDFilterSweetSpotDataset
    
    sweet_spot_dataset = create_nd_sweet_spot_dataset(
        aggregated_sweet_spot_data=combined_sweet_spot_data,
        common_metadata=common_metadata, od_var_flag = od_var_flag, input_dataset = stars_dataset
    )

    #TO DO: do we want to return flux?
    return sweet_spot_dataset


"""
Module to hold input-checking functions to minimize repetition

Copied over from the II&T pipeline
"""
import numbers

import numpy as np

class CheckException(Exception):
    pass

# String check support
string_types = (str, bytes)

# Int check support
int_types = (int, np.integer)

def _checkname(vname):
    """
    Internal check that we can use vname as a string for printing

    Args:
    vname: str
        variable name
    """
    if not isinstance(vname, string_types):
        raise CheckException('vname must be a string when fed to check ' + \
                             'functions')
    pass


def _checkexc(vexc):
    """
    Internal check that we can raise from the vexc object

    Args:
    vexc: str
        exception name
    """
    if not isinstance(vexc, type): # pre-check it is class-like
        raise CheckException('vexc must be a Exception, or an object ' + \
                             'descended from one when fed to check functions')
    if not issubclass(vexc, Exception):
        raise CheckException('vexc must be a Exception, or an object ' + \
                             'descended from one when fed to check functions')
    pass


def real_positive_scalar(var, vname, vexc):
    """
    Checks whether an object is a real positive scalar.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    if not isinstance(var, numbers.Number):
        raise vexc(vname + ' must be scalar')
    if not np.isrealobj(var):
        raise vexc(vname + ' must be real')
    if var <= 0:
        raise vexc(vname + ' must be positive')
    return var


def real_array(var, vname, vexc):
    """
    Checks whether an object is a real numpy array, or castable to one.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    var = np.asarray(var) # cast to array
    if len(var.shape) == 0:
        raise vexc(vname + ' must have length > 0')
    if not np.isrealobj(var):
        raise vexc(vname + ' must be a real array')
    # skip 'c' as we don't want complex; rest are non-numeric
    if not var.dtype.kind in ['b', 'i', 'u', 'f']:
        raise vexc(vname + ' must be a real numeric type to be real')
    return var


def oneD_array(var, vname, vexc):
    """
    Checks whether an object is a 1D numpy array, or castable to one.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    var = np.asarray(var) # cast to array
    if len(var.shape) != 1:
        raise vexc(vname + ' must be a 1D array')
    if (not np.isrealobj(var)) and (not np.iscomplexobj(var)):
        raise vexc(vname + ' must be a real or complex 1D array')
    return var


def twoD_array(var, vname, vexc):
    """
    Checks whether an object is a 2D numpy array, or castable to one.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    var = np.asarray(var) # cast to array
    if len(var.shape) != 2:
        raise vexc(vname + ' must be a 2D array')
    if (not np.isrealobj(var)) and (not np.iscomplexobj(var)):
        raise vexc(vname + ' must be a real or complex 2D array')
    return var


def twoD_square_array(var, vname, vexc):
    """
    Checks whether an object is a 2D square array_like.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    var = np.asarray(var) # cast to array
    if len(var.shape) != 2:
        raise vexc(vname + ' must be a 2D array')
    else: # is 2-D
        if not var.shape[0] == var.shape[1]:
            raise vexc(vname + ' must be a square 2D array')
    if (not np.isrealobj(var)) and (not np.iscomplexobj(var)):
        raise vexc(vname + ' must be a real or complex square 2D array')
    return var


def threeD_array(var, vname, vexc):
    """
    Checks whether an object is a 3D numpy array, or castable to one.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    var = np.asarray(var) # cast to array
    if len(var.shape) != 3:
        raise vexc(vname + ' must be a 3D array')
    if (not np.isrealobj(var)) and (not np.iscomplexobj(var)):
        raise vexc(vname + ' must be a real or complex 3D array')
    return var



def real_scalar(var, vname, vexc):
    """
    Checks whether an object is a real scalar.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    if not isinstance(var, numbers.Number):
        raise vexc(vname + ' must be scalar')
    if not np.isrealobj(var):
        raise vexc(vname + ' must be real')
    return var


def real_nonnegative_scalar(var, vname, vexc):
    """
    Checks whether an object is a real nonnegative scalar.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    if not isinstance(var, numbers.Number):
        raise vexc(vname + ' must be scalar')
    if not np.isrealobj(var):
        raise vexc(vname + ' must be real')
    if var < 0:
        raise vexc(vname + ' must be nonnegative')
    return var


def positive_scalar_integer(var, vname, vexc):
    """
    Checks whether an object is a positive scalar integer.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    if not isinstance(var, numbers.Number):
        raise vexc(vname + ' must be scalar')
    if not isinstance(var, int_types):
        raise vexc(vname + ' must be integer')
    if var <= 0:
        raise vexc(vname + ' must be positive')
    return var


def nonnegative_scalar_integer(var, vname, vexc):
    """
    Checks whether an object is a nonnegative scalar integer.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    if not isinstance(var, numbers.Number):
        raise vexc(vname + ' must be scalar')
    if not isinstance(var, int_types):
        raise vexc(vname + ' must be integer')
    if var < 0:
        raise vexc(vname + ' must be nonnegative')
    return var


def scalar_integer(var, vname, vexc):
    """
    Checks whether an object is a scalar integer (no sign dependence).

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    if not isinstance(var, numbers.Number):
        raise vexc(vname + ' must be scalar')
    if not isinstance(var, int_types):
        raise vexc(vname + ' must be integer')
    return var


def string(var, vname, vexc):
    """
    Checks whether an object is a string.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    if not isinstance(var, string_types):
        raise vexc(vname + ' must be a string')
    return var


def boolean(var, vname, vexc):
    """
    Checks whether an object is a bool.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    if not isinstance(var, bool):
        raise vexc(vname + ' must be a bool')
    return var


def dictionary(var, vname, vexc):
    """
    Checks whether an object is a dictionary.

    Arguments:
     var: variable to check
     vname: string to output in case of error for debugging
     vexc: Exception to raise in case of error for debugging

    Returns:
     returns var

    """
    _checkname(vname)
    _checkexc(vexc)

    if not isinstance(var, dict):
        raise vexc(vname + ' must be a dict')
    return var
